import copy
import random
import string
import sys
from argparse import _UNRECOGNIZED_ARGS_ATTR, PARSER, REMAINDER, SUPPRESS, Action, ArgumentError
from argparse import ArgumentParser as ArgParser
from argparse import Namespace, _SubParsersAction, _VersionAction
from typing import IO, Any, Callable, List, Optional, TextIO

import flexget
from flexget.entry import Entry
from flexget.event import fire_event
from flexget.utils.tools import get_current_flexget_version, get_latest_flexget_version_number

_UNSET = object()

core_parser: Optional['CoreArgumentParser'] = None


def unicode_argv() -> List[str]:
    """Like sys.argv, but decodes all arguments."""
    args = []
    for arg in sys.argv:
        if isinstance(arg, bytes):
            arg = arg.decode(sys.getfilesystemencoding())
        args.append(arg)
    return args


def get_parser(command: str = None) -> 'ArgumentParser':
    global core_parser
    if not core_parser:
        core_parser = CoreArgumentParser()
        # Add all plugin options to the parser
        fire_event('options.register')
    if command:
        return core_parser.get_subparser(command)
    return core_parser


def register_command(
    command: str, callback: Callable[['flexget.manager.Manager', Namespace], Any], **kwargs
) -> 'ArgumentParser':
    """
    Register a callback function to be executed when flexget is launched with the given `command`.

    :param command: The command being defined.
    :param callback: Callback function executed when this command is invoked from the CLI. Should take manager instance
        and parsed argparse namespace as parameters.
    :param kwargs: Other keyword arguments will be passed to the :class:`arparse.ArgumentParser` constructor
    :returns: An :class:`argparse.ArgumentParser` instance ready to be configured with the options for this command.
    """
    return get_parser().add_subparser(
        command, parent_defaults={'cli_command_callback': callback}, **kwargs
    )


def required_length(nmin: int, nmax: int):
    """Generates a custom Action to validate an arbitrary range of arguments."""

    class RequiredLength(Action):
        def __call__(self, parser: ArgParser, args, values, option_string=None):
            if not nmin <= len(values) <= nmax:
                raise ArgumentError(self, 'requires between %s and %s arguments' % (nmin, nmax))
            setattr(args, self.dest, values)

    return RequiredLength


class VersionAction(_VersionAction):
    """Action to print the current version. Also checks latest release revision."""

    def __call__(self, parser: ArgParser, namespace: Namespace, values, option_string=None):
        from flexget.terminal import console

        current = get_current_flexget_version()
        latest = get_latest_flexget_version_number()

        # Print the version number
        console('%s' % get_current_flexget_version())
        # Check for latest version from server
        if latest:
            if current == latest:
                console('You are on the latest release.')
            else:
                console('Latest release: %s' % latest)
        else:
            console(
                'Error getting latest version number from https://pypi.python.org/pypi/FlexGet'
            )
        parser.exit()


class HelpAction(Action):
    """Override the default help command so that we can conditionally disable it to prevent program exit."""

    def __call__(self, parser, namespace, values, option_string=None):
        if getattr(parser, 'do_help', True):
            parser.print_help()
            parser.exit()


class DebugAction(Action):
    def __call__(self, parser, namespace, values, option_string=None):
        setattr(namespace, self.dest, True)
        namespace.loglevel = 'DEBUG'


class DebugTraceAction(Action):
    def __call__(self, parser, namespace, values, option_string=None):
        setattr(namespace, self.dest, True)
        namespace.debug = True
        namespace.log_level = 'trace'


class CronAction(Action):
    def __call__(self, parser, namespace, values, option_string=None):
        setattr(namespace, self.dest, True)
        # Only set loglevel if it has not already explicitly been set
        if not hasattr(namespace, 'loglevel'):
            namespace.loglevel = 'INFO'


# This makes the old --inject form forwards compatible
class InjectAction(Action):
    def __call__(self, parser, namespace, values, option_string=None):
        kwargs = {'title': values.pop(0)}
        if values:
            kwargs['url'] = values.pop(0)
        else:
            kwargs['url'] = 'http://localhost/inject/%s' % ''.join(
                random.sample(string.ascii_letters + string.digits, 30)
            )
        if 'force' in [v.lower() for v in values]:
            kwargs['immortal'] = True
        entry = Entry(**kwargs)
        if 'accept' in [v.lower() for v in values]:
            entry.accept(reason='accepted by --inject')
        existing = getattr(namespace, self.dest, None) or []
        setattr(namespace, self.dest, existing + [entry])


class ParseExtrasAction(Action):
    """This action will take extra arguments, and parser them with a different parser."""

    def __init__(self, option_strings, parser, help=None, metavar=None, dest=None, required=False):
        if metavar is None:
            metavar = '<%s arguments>' % parser.prog
        if help is None:
            help = 'arguments for the `%s` command are allowed here' % parser.prog
        self._parser = parser
        super().__init__(
            option_strings=option_strings,
            dest=SUPPRESS,
            help=help,
            metavar=metavar,
            nargs=REMAINDER,
            required=required,
        )

    def __call__(self, parser, namespace, values, option_string=None):
        namespace, extras = self._parser.parse_known_args(values, namespace)
        if extras:
            parser.error('unrecognized arguments: %s' % ' '.join(extras))


class ScopedNamespace(Namespace):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.__parent__ = None

    def __getattr__(self, key):
        if '.' in key:
            scope, key = key.split('.', 1)
            return getattr(getattr(self, scope), key)

        if self.__parent__:
            return getattr(self.__parent__, key)
        raise AttributeError("'%s' object has no attribute '%s'" % (type(self).__name__, key))

    def __setattr__(self, key: str, value):
        if '.' in key:
            scope, key = key.split('.', 1)
            if not hasattr(self, scope):
                setattr(self, scope, type(self)())
            sub_ns = getattr(self, scope, None)
            return object.__setattr__(sub_ns, key, value)
        # Let child namespaces keep track of us
        if key != '__parent__' and isinstance(value, ScopedNamespace):
            value.__parent__ = self
        return object.__setattr__(self, key, value)

    def __iter__(self):
        return (i for i in self.__dict__.items() if i[0] != '__parent__')

    def __copy__(self):
        new = self.__class__()
        new.__dict__.update(self.__dict__)
        # Make copies of any nested namespaces
        for key, value in self:
            if isinstance(value, ScopedNamespace):
                setattr(new, key, copy.copy(value))
        return new


class NestedSubparserAction(_SubParsersAction):
    def __init__(self, *args, **kwargs):
        self.nested_namespaces = kwargs.pop('nested_namespaces', False)
        self.parent_defaults = {}
        # Python < 3.7 doesn't support the 'required' argument for subparsers
        required = kwargs.pop('required', True)
        super().__init__(*args, **kwargs)
        self.required = required

    def add_parser(self, name: str, parent_defaults: dict = None, **kwargs):
        if parent_defaults:
            self.parent_defaults[name] = parent_defaults
        return super().add_parser(name, **kwargs)

    def __call__(self, parser, namespace, values, option_string=None):
        parser_name = values[0]
        if parser_name in self.parent_defaults:
            for dest in self.parent_defaults[parser_name]:
                if not hasattr(namespace, dest):
                    setattr(namespace, dest, self.parent_defaults[parser_name][dest])
        if self.nested_namespaces:
            subnamespace = ScopedNamespace()
            super().__call__(parser, subnamespace, values, option_string)
            # If dest is set, it should be set on the parent namespace, not subnamespace
            if self.dest is not SUPPRESS:
                setattr(namespace, self.dest, parser_name)
                delattr(subnamespace, self.dest)
            setattr(namespace, parser_name, subnamespace)
            # Propagate unrecognized arguments back to parent namespace
            vars(namespace).setdefault(_UNRECOGNIZED_ARGS_ATTR, [])
            getattr(namespace, _UNRECOGNIZED_ARGS_ATTR).extend(
                getattr(subnamespace, _UNRECOGNIZED_ARGS_ATTR)
            )
        else:
            super().__call__(parser, namespace, values, option_string)


class ParserError(Exception):
    def __init__(self, message, parser):
        self.message = message
        self.parser = parser

    def __unicode__(self):
        return self.message

    def __repr__(self):
        return 'ParserError(%s, %s)' % (self.message, self.parser)


class ArgumentParser(ArgParser):
    """
    Mimics the default :class:`argparse.ArgumentParser` class, with a few distinctions, mostly to ease subparser usage:

    - If `add_subparsers` is called with the `nested_namespaces` kwarg, all subcommand options will be stored in a
      nested namespace based on the command name for the subparser
    - Adds the `add_subparser` method. After `add_subparsers` has been called, the `add_subparser` method can be used
      instead of the `add_parser` method of the object returned by the `add_subparsers` call.
    - `add_subparser` takes takes the `parent_defaults` argument, which will set/change the defaults for the parent
      parser when that subparser is selected.
    - The `get_subparser` method will get the :class:`ArgumentParser` instance for an existing subparser on this parser
    - For any arguments defined both in this parser and one of its subparsers, the selected subparser default will
      override the main one.
    - Adds the `set_post_defaults` method. This works like the normal argparse `set_defaults` method, but all actions
      and subparsers will be run before any of these defaults are set.
    - Command shortening: If the command for a subparser is abbreviated unambiguously, it will still be accepted.
    - The add_argument `nargs` keyword argument supports a range of arguments, e.g. `"2-4"
    - If the `raise_errors` keyword argument to `parse_args` is True, a `ParserError` will be raised instead of sys.exit
    - If the `file` argument is given to `parse_args`, output will be printed there instead of sys.stdout or stderr
    """

    # These are created as a class attribute so that we can set it for parser and all subparsers at once
    file: Optional[IO[str]] = None
    do_help = True

    def __init__(self, **kwargs):
        """
        :param nested_namespace_name: When used as a subparser, options from this parser will be stored nested under
            this attribute name in the root parser's namespace
        """
        # Do this early, so even option processing stuff is caught
        if '--bugreport' in unicode_argv():
            self._debug_tb_callback()

        self.subparsers = None
        self.raise_errors = None
        add_help = kwargs.pop('add_help', True)
        kwargs['add_help'] = False
        ArgParser.__init__(self, **kwargs)
        if add_help:
            self.add_argument(
                '--help',
                '-h',
                action=HelpAction,
                dest=SUPPRESS,
                default=SUPPRESS,
                nargs=0,
                help='Show this help message and exit',
            )
        # Overwrite _SubparserAction with our custom one
        self.register('action', 'parsers', NestedSubparserAction)

        self.post_defaults = {}
        if kwargs.get('parents'):
            for parent in kwargs['parents']:
                if hasattr(parent, 'post_defaults'):
                    self.set_post_defaults(**parent.post_defaults)

    def add_argument(self, *args, **kwargs):
        if isinstance(kwargs.get('nargs'), str) and '-' in kwargs['nargs']:
            # Handle a custom range of arguments
            min, max = kwargs['nargs'].split('-')
            min, max = int(min), int(max)
            kwargs['action'] = required_length(min, max)
            # Make the usage string a bit better depending on whether the first argument is optional
            if min == 0:
                kwargs['nargs'] = '*'
            else:
                kwargs['nargs'] = '+'
        return super().add_argument(*args, **kwargs)

    def _print_message(self, message, file=None):
        """If a file argument was passed to `parse_args` make sure output goes there."""
        if self.file:
            file = self.file
        super()._print_message(message, file)

    def set_post_defaults(self, **kwargs):
        """Like set_defaults method, but these defaults will be defined after parsing instead of before."""
        self.post_defaults.update(kwargs)

        # if these defaults match any existing arguments, suppress
        # the previous default so that it can be filled after parsing
        for action in self._actions:
            if action.dest in kwargs:
                action.default = SUPPRESS

    def error(self, msg: str):
        raise ParserError(msg, self)

    def parse_args(
        self,
        args: List[str] = None,
        namespace: Namespace = None,
        raise_errors: bool = False,
        file: TextIO = None,
    ):  # pylint: disable=W0221
        """
        :param raise_errors: If this is true, errors will be raised as `ParserError`s instead of calling sys.exit
        """
        ArgumentParser.file = file
        try:
            return super().parse_args(args, namespace)
        except ParserError as e:
            if raise_errors:
                raise
            super(ArgumentParser, e.parser).error(e.message)
        finally:
            ArgumentParser.file = None

    def parse_known_args(
        self,
        args: List[str] = None,
        namespace: Namespace = None,
        do_help: bool = None,
    ):
        if args is None:
            # Decode all arguments to unicode before parsing
            args = unicode_argv()[1:]
        if namespace is None:
            namespace = ScopedNamespace()
        old_do_help = ArgumentParser.do_help
        if do_help is not None:
            ArgumentParser.do_help = do_help
        try:
            namespace, args = super().parse_known_args(args, namespace)
        finally:
            ArgumentParser.do_help = old_do_help

        # add any post defaults that aren't present
        for dest in self.post_defaults:
            if not hasattr(namespace, dest):
                setattr(namespace, dest, self.post_defaults[dest])

        return namespace, args

    def add_subparsers(self, **kwargs):
        """
        :param nested_namespaces: If True, options from subparsers will appear in nested namespace under the subparser
            name.
        """
        # Set the parser class so subparsers don't end up being an instance of a subclass, like CoreArgumentParser
        kwargs.setdefault('parser_class', ArgumentParser)
        kwargs.setdefault('required', True)
        self.subparsers = super().add_subparsers(**kwargs)
        return self.subparsers

    def add_subparser(self, name: str, **kwargs):
        """
        Adds a parser for a new subcommand and returns it.

        :param name: Name of the subcommand
        :param parent_defaults: Default argument values which should be supplied to the parent parser if this subparser
            is selected.
        """
        if not self.subparsers:
            raise TypeError('This parser does not have subparsers')
        result = self.subparsers.add_parser(name, **kwargs)
        return result

    def get_subparser(self, name: str, default=_UNSET):
        if not self.subparsers:
            raise TypeError('This parser does not have subparsers')
        p = self.subparsers.choices.get(name, default)
        if p is _UNSET:
            raise ValueError('%s is not an existing subparser name' % name)
        return p

    def _get_values(self, action, arg_strings):
        """Complete the full name for partial subcommands"""
        if action.nargs == PARSER and self.subparsers:
            subcommand = arg_strings[0]
            if subcommand not in self.subparsers.choices:
                matches = [x for x in self.subparsers.choices if x.startswith(subcommand)]
                if len(matches) == 1:
                    arg_strings[0] = matches[0]
        return super()._get_values(action, arg_strings)

    def _debug_tb_callback(self, *dummy):
        import cgitb

        cgitb.enable(format="text")


# This will hold just the arguments directly for Manager.
manager_parser = ArgumentParser(add_help=False)
manager_parser.add_argument(
    '-V',
    '--version',
    action=VersionAction,
    version=flexget.__version__,
    help='Print FlexGet version and exit.',
)
manager_parser.add_argument(
    '--test',
    action='store_true',
    dest='test',
    default=0,
    help='Verbose what would happen on normal execution.',
)
manager_parser.add_argument(
    '-c',
    dest='config',
    default='config.yml',
    help='Specify configuration file. Default: %(default)s',
)
manager_parser.add_argument(
    '--logfile',
    '-l',
    default='flexget.log',
    help='Specify a custom logfile name/location. '
    'Default: %(default)s in the config directory.',
)
manager_parser.add_argument(
    '--loglevel',
    '-L',
    metavar='LEVEL',
    help='Set the verbosity of the logger. Levels: %(choices)s',
    choices=['NONE', 'CRITICAL', 'ERROR', 'WARNING', 'INFO', 'VERBOSE', 'DEBUG', 'TRACE'],
    type=str.upper,
)
manager_parser.set_post_defaults(loglevel='VERBOSE')
# This option is already handled above.
manager_parser.add_argument(
    '--bugreport',
    action='store_true',
    dest='debug_tb',
    help='Use this option to create a detailed bug report, '
    'note that the output might contain PRIVATE data, so edit that out',
)
manager_parser.add_argument(
    '--profile',
    metavar='OUTFILE',
    nargs='?',
    const='flexget.profile',
    help='Use the python profiler for this run to debug performance issues.',
)
manager_parser.add_argument('--debug', action=DebugAction, nargs=0, help=SUPPRESS)
manager_parser.add_argument('--debug-trace', action=DebugTraceAction, nargs=0, help=SUPPRESS)
manager_parser.add_argument('--debug-sql', action='store_true', default=False, help=SUPPRESS)
manager_parser.add_argument('--experimental', action='store_true', default=False, help=SUPPRESS)
manager_parser.add_argument('--ipc-port', type=int, help=SUPPRESS)
manager_parser.add_argument(
    '--cron',
    action=CronAction,
    default=False,
    nargs=0,
    help='use when executing FlexGet non-interactively: allows background '
    'maintenance to run, disables stdout and stderr output, reduces logging level',
)


class CoreArgumentParser(ArgumentParser):
    """
    The core argument parser, contains the manager arguments, command parsers, and plugin arguments.

    Warning: Only gets plugin arguments if instantiated after plugins have been loaded.

    """

    def __init__(self, **kwargs):
        kwargs.setdefault('parents', [manager_parser])
        kwargs.setdefault('prog', 'flexget')
        super().__init__(**kwargs)
        self.add_subparsers(
            title='commands', metavar='<command>', dest='cli_command', nested_namespaces=True
        )

        # The parser for the execute command
        exec_parser = self.add_subparser('execute', help='execute tasks now')
        exec_parser.add_argument(
            '--tasks',
            nargs='+',
            metavar='TASK',
            help='run only specified task(s), optionally using glob patterns ("tv-*"). '
            'matching is case-insensitive',
        )
        exec_parser.add_argument(
            '--learn',
            action='store_true',
            dest='learn',
            default=False,
            help='matches are not downloaded but will be skipped in the future',
        )
        exec_parser.add_argument('--profile', action='store_true', default=False, help=SUPPRESS)
        exec_parser.add_argument('--disable-phases', nargs='*', help=SUPPRESS)
        exec_parser.add_argument('--inject', nargs='+', action=InjectAction, help=SUPPRESS)
        # Plugins should respect these flags where appropriate
        exec_parser.add_argument(
            '--retry', action='store_true', dest='retry', default=False, help=SUPPRESS
        )
        exec_parser.add_argument(
            '--no-cache',
            action='store_true',
            dest='nocache',
            default=False,
            help='disable caches. works only in plugins that have explicit support',
        )

        daemonize_help = SUPPRESS
        if not sys.platform.startswith('win'):
            daemonize_help = 'causes process to daemonize after starting'

        # The parser for the daemon command
        daemon_parser = self.add_subparser(
            'daemon',
            parent_defaults={'loglevel': 'INFO'},
            help='run continuously, executing tasks according to schedules defined in config',
        )
        daemon_parser.add_subparsers(title='actions', metavar='<action>', dest='action')
        start_parser = daemon_parser.add_subparser('start', help='start the daemon')
        start_parser.add_argument('-d', '--daemonize', action='store_true', help=daemonize_help)
        start_parser.add_argument(
            '--autoreload-config',
            action='store_true',
            help='automatically reload the config from disk if the daemon detects any changes',
        )
        start_parser.add_argument(
            '--enable-tray-icon',
            action='store_true',
            dest='tray_icon',
            help='Enable the tray icon',
        )
        stop_parser = daemon_parser.add_subparser('stop', help='shutdown the running daemon')
        stop_parser.add_argument(
            '--wait',
            action='store_true',
            help='wait for all queued tasks to finish before stopping daemon',
        )
        daemon_parser.add_subparser('status', help='check if a daemon is running')
        daemon_parser.add_subparser(
            'reload-config', help='causes a running daemon to reload the config from disk'
        )

    def add_subparsers(self, **kwargs):
        # The subparsers should not be CoreArgumentParsers
        kwargs.setdefault('parser_class', ArgumentParser)
        return super().add_subparsers(**kwargs)

    def parse_args(self, *args, **kwargs):
        result = super().parse_args(*args, **kwargs)
        # Make sure we always have execute parser settings even when other commands called
        if not result.cli_command == 'execute':
            exec_options = get_parser('execute').parse_args([])
            if hasattr(result, 'execute'):
                exec_options.__dict__.update(result.execute.__dict__)
            result.execute = exec_options
        # Set the 'allow_manual' flag to True for any usage of the CLI
        setattr(result, 'allow_manual', True)
        return result



import datetime
import itertools
import random

from loguru import logger
from sqlalchemy import Column, DateTime, Index, Integer, Unicode

from flexget import db_schema, options, plugin
from flexget.event import event
from flexget.manager import Session
from flexget.utils.tools import aggregate_inputs, multiply_timedelta, parse_timedelta

logger = logger.bind(name='discover')
Base = db_schema.versioned_base('discover', 0)


class DiscoverEntry(Base):
    __tablename__ = 'discover_entry'

    id = Column(Integer, primary_key=True)
    title = Column(Unicode, index=True)
    task = Column(Unicode, index=True)
    last_execution = Column(DateTime)

    def __init__(self, title, task):
        self.title = title
        self.task = task
        self.last_execution = None

    def __str__(self):
        return '<DiscoverEntry(title=%s,task=%s,added=%s)>' % (
            self.title,
            self.task,
            self.last_execution,
        )


Index('ix_discover_entry_title_task', DiscoverEntry.title, DiscoverEntry.task)


@event('manager.db_cleanup')
def db_cleanup(manager, session):
    value = datetime.datetime.now() - parse_timedelta('7 days')
    for discover_entry in (
        session.query(DiscoverEntry).filter(DiscoverEntry.last_execution <= value).all()
    ):
        logger.debug('deleting {}', discover_entry)
        session.delete(discover_entry)


class Discover:
    """
    Discover content based on other inputs material.

    Example::

      discover:
        what:
          - next_series_episodes: yes
        from:
          - piratebay
        interval: [1 hours|days|weeks]
        release_estimations: [strict|loose|ignore]
    """

    schema = {
        'type': 'object',
        'properties': {
            'what': {
                'type': 'array',
                'items': {
                    'allOf': [
                        {'$ref': '/schema/plugins?phase=input'},
                        {'maxProperties': 1, 'minProperties': 1},
                    ]
                },
            },
            'from': {
                'type': 'array',
                'items': {
                    'allOf': [
                        {'$ref': '/schema/plugins?interface=search'},
                        {'maxProperties': 1, 'minProperties': 1},
                    ]
                },
            },
            'interval': {'type': 'string', 'format': 'interval', 'default': '5 hours'},
            'release_estimations': {
                'oneOf': [
                    {
                        'type': 'string',
                        'default': 'strict',
                        'enum': ['loose', 'strict', 'ignore', 'smart'],
                    },
                    {
                        'type': 'object',
                        'properties': {'optimistic': {'type': 'string', 'format': 'interval'}},
                        'required': ['optimistic'],
                    },
                ]
            },
            'limit': {'type': 'integer', 'minimum': 1},
        },
        'required': ['what', 'from'],
        'additionalProperties': False,
    }

    def execute_searches(self, config, entries, task):
        """
        :param config: Discover plugin config
        :param entries: List of pseudo entries to search
        :param task: Task being run
        :return: List of entries found from search engines listed under `from` configuration
        """

        result = []
        for index, entry in enumerate(entries):
            entry_results = []
            for item in config['from']:
                if isinstance(item, dict):
                    plugin_name, plugin_config = list(item.items())[0]
                else:
                    plugin_name, plugin_config = item, None
                search = plugin.get(plugin_name, self)
                if not callable(getattr(search, 'search')):
                    logger.critical(
                        'Search plugin {} does not implement search method', plugin_name
                    )
                    continue
                logger.verbose(
                    'Searching for `{}` with plugin `{}` ({} of {})',
                    entry['title'],
                    plugin_name,
                    index + 1,
                    len(entries),
                )
                try:
                    search_results = search.search(task=task, entry=entry, config=plugin_config)
                    if not search_results:
                        logger.debug('No results from {}', plugin_name)
                        continue
                    if config.get('limit'):
                        search_results = itertools.islice(search_results, config['limit'])
                    # 'search_results' can be any iterable, make sure it's a list.
                    search_results = list(search_results)
                    logger.debug('Discovered {} entries from {}', len(search_results), plugin_name)
                    for e in search_results:
                        e['discovered_from'] = entry['title']
                        e['discovered_with'] = plugin_name
                        e.on_complete(
                            self.entry_complete, query=entry, search_results=search_results
                        )

                    entry_results.extend(search_results)

                except plugin.PluginWarning as e:
                    logger.verbose('No results from {}: {}', plugin_name, e)
                except plugin.PluginError as e:
                    logger.error('Error searching with {}: {}', plugin_name, e)
            if not entry_results:
                logger.verbose('No search results for `{}`', entry['title'])
                entry.complete()
                continue
            result.extend(entry_results)

        return result

    def entry_complete(self, entry, query=None, search_results=None, **kwargs):
        """Callback for Entry"""
        if entry.accepted:
            # One of the search results was accepted, transfer the acceptance back to the query entry which generated it
            query.accept()
        # Remove this entry from the list of search results yet to complete
        search_results.remove(entry)
        # When all the search results generated by a query entry are complete, complete the query which generated them
        if not search_results:
            query.complete()

    def estimated(self, entries, estimation_mode):
        """
        :param dict estimation_mode: mode -> loose, strict, ignore
        :return: Entries that we have estimated to be available
        """
        estimator = plugin.get('estimate_release', self)
        result = []
        for entry in entries:
            estimation = estimator.estimate(entry)

            est_date = estimation['entity_date']
            data_exists = estimation['data_exists']

            if est_date is None:
                if estimation_mode['mode'] == 'strict':
                    logger.debug('No release date could be determined for {}', entry['title'])
                    entry.reject('has no release date')
                    entry.complete()
                elif estimation_mode['mode'] == 'smart' and data_exists:
                    logger.debug(
                        'No release date could be determined for {}, but exists data',
                        entry['title'],
                    )
                    entry.reject('exists but has no release date')
                    entry.complete()
                elif estimation_mode['mode'] == 'smart' and not data_exists:
                    logger.debug(
                        'Discovering because mode is \'{}\' and no data is found for entry',
                        estimation_mode['mode'],
                    )
                    result.append(entry)
                else:
                    result.append(entry)
                continue
            if isinstance(est_date, datetime.date):
                # If we just got a date, add a time so we can compare it to now()
                est_date = datetime.datetime.combine(est_date, datetime.time())
            if datetime.datetime.now() >= est_date:
                logger.debug('{} has been released at {}', entry['title'], est_date)
                result.append(entry)
            elif datetime.datetime.now() >= est_date - parse_timedelta(
                estimation_mode['optimistic']
            ):
                logger.debug(
                    '{} will be released at {}. Ignoring release estimation because estimated release date is in less than {}',
                    entry['title'],
                    est_date,
                    estimation_mode['optimistic'],
                )
                result.append(entry)
            else:
                entry.reject('has not been released')
                entry.complete()
                logger.verbose(
                    "{} hasn't been released yet (Expected: {})", entry['title'], est_date
                )
        return result

    def interval_expired(self, config, task, entries):
        """
        Maintain some limit levels so that we don't hammer search
        sites with unreasonable amount of queries.

        :return: Entries that are up for ``config['interval']``
        """
        config.setdefault('interval', '5 hour')
        interval = parse_timedelta(config['interval'])
        if task.options.discover_now:
            logger.info('Ignoring interval because of --discover-now')
        result = []
        interval_count = 0
        with Session() as session:
            for entry in entries:
                discover_entry = (
                    session.query(DiscoverEntry)
                    .filter(DiscoverEntry.title == entry['title'])
                    .filter(DiscoverEntry.task == task.name)
                    .first()
                )

                if not discover_entry:
                    logger.debug('{} -> No previous run recorded', entry['title'])
                    discover_entry = DiscoverEntry(entry['title'], task.name)
                    session.add(discover_entry)
                if (
                    not task.is_rerun and task.options.discover_now
                ) or not discover_entry.last_execution:
                    # First time we execute (and on --discover-now) we randomize time to avoid clumping
                    delta = multiply_timedelta(interval, random.random())
                    discover_entry.last_execution = datetime.datetime.now() - delta
                else:
                    next_time = discover_entry.last_execution + interval
                    logger.debug(
                        'last_time: {!r}, interval: {}, next_time: {!r}, ',
                        discover_entry.last_execution,
                        config['interval'],
                        next_time,
                    )
                    if datetime.datetime.now() < next_time:
                        logger.debug('interval not met')
                        interval_count += 1
                        entry.reject('discover interval not met')
                        entry.complete()
                        continue
                    discover_entry.last_execution = datetime.datetime.now()
                logger.trace('interval passed for {}', entry['title'])
                result.append(entry)
        if interval_count and not task.is_rerun:
            logger.verbose(
                'Discover interval of {} not met for {} entries. Use --discover-now to override.',
                config['interval'],
                interval_count,
            )
        return result

    def on_task_input(self, task, config):
        config.setdefault('release_estimations', {})
        if not isinstance(config['release_estimations'], dict):
            config['release_estimations'] = {'mode': config['release_estimations']}

        config['release_estimations'].setdefault('mode', 'strict')
        config['release_estimations'].setdefault('optimistic', '0 days')

        task.no_entries_ok = True
        entries = aggregate_inputs(task, config['what'])
        logger.verbose('Discovering {} titles ...', len(entries))
        if len(entries) > 500:
            logger.critical(
                'Looks like your inputs in discover configuration produced '
                'over 500 entries, please reduce the amount!'
            )
        # TODO: the entries that are estimated should be given priority over expiration
        entries = self.interval_expired(config, task, entries)
        estimation_mode = config['release_estimations']
        if estimation_mode['mode'] != 'ignore':
            entries = self.estimated(entries, estimation_mode)
        return self.execute_searches(config, entries, task)


@event('plugin.register')
def register_plugin():
    plugin.register(Discover, 'discover', api_ver=2)


@event('options.register')
def register_parser_arguments():
    options.get_parser('execute').add_argument(
        '--discover-now',
        action='store_true',
        dest='discover_now',
        default=False,
        help='Immediately try to discover everything',
    )



import pytest

from flexget.components.imdb.api import ObjectsContainer as OC
from flexget.utils import json


@pytest.mark.online
class TestIMDBLookupAPI:
    config = 'tasks: {}'

    def test_imdb_search(self, api_client, schema_match):
        # No params
        rsp = api_client.get('/imdb/search/')
        assert rsp.status_code == 404, 'Response code is %s' % rsp.status_code

        # Lookup by title
        rsp = api_client.get('/imdb/search/saw/')
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code

        data = json.loads(rsp.get_data(as_text=True))
        errors = schema_match(OC.return_object, data)
        assert not errors
        assert len(data) > 1

        # Lookup by IMDB ID
        rsp = api_client.get('/imdb/search/tt0234215/')
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code

        data = json.loads(rsp.get_data(as_text=True))
        errors = schema_match(OC.return_object, data)
        assert not errors

        assert len(data) == 1

        # Lookup non-existing title
        rsp = api_client.get('/imdb/search/sdfgsdfgsdfgsdfg/')
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code

        data = json.loads(rsp.get_data(as_text=True))
        errors = schema_match(OC.return_object, data)
        assert not errors

        assert len(data) == 0



from loguru import logger

from flexget import plugin
from flexget.event import event
from flexget.utils import bittorrent

try:
    # NOTE: Importing other plugins is discouraged!
    from flexget.components.bittorrent import torrent as plugin_torrent
except ImportError:
    raise plugin.DependencyError(issued_by=__name__, missing='torrent')

logger = logger.bind(name='torrent_scrub')


class TorrentScrub:
    """Scrubs torrents from unwanted keys.

    Example:
        tasks:
          rutorrent-fast-resume-infected-task:
            torrent_scrub: resume
    """

    # Scrub at high level, but BELOW "torrent"
    SCRUB_PRIO = plugin_torrent.TorrentFilename.TORRENT_PRIO - 10

    # Scrubbing modes
    SCRUB_MODES = ("off", "on", "all", "resume", "rtorrent")

    # Keys of rTorrent / ruTorrent session data
    RT_KEYS = ("libtorrent_resume", "log_callback", "err_callback", "rtorrent")

    schema = {
        'oneOf': [
            {'type': 'boolean'},
            {'type': 'string', 'enum': list(SCRUB_MODES)},
            {'type': 'array', 'items': {'type': 'string'}},  # list of keys to scrub
        ]
    }

    @plugin.priority(SCRUB_PRIO)
    def on_task_modify(self, task, config):
        """Scrub items that are torrents, if they're affected."""
        if isinstance(config, list):
            mode = "fields"
        else:
            mode = str(config).lower()
            if mode in ("off", "false"):
                logger.debug("Plugin configured, but disabled")
                return

        for entry in task.entries:
            # Skip non-torrents
            if "torrent" not in entry:
                continue

            # Scrub keys as configured
            modified = set()
            metainfo = entry["torrent"].content
            infohash = entry["torrent"].info_hash

            if mode in ("on", "all", "true"):
                modified = bittorrent.clean_meta(
                    metainfo, including_info=(mode == "all"), log_func=logger.debug
                )
            elif mode in ("resume", "rtorrent"):
                if mode == "resume":
                    self.RT_KEYS = self.RT_KEYS[:1]

                for key in self.RT_KEYS:
                    if key in metainfo:
                        logger.debug("Removing key '{}'...", key)
                        del metainfo[key]
                        modified.add(key)
            elif mode == "fields":
                # Scrub all configured fields
                for key in config:
                    fieldname = key  # store for logging
                    key = bittorrent.Torrent.KEY_TYPE(key)
                    field = metainfo

                    while field and '.' in key:
                        name, key = key.split('.', 1)
                        try:
                            field = field[name]
                        except KeyError:
                            # Key not found in this entry
                            field = None
                        logger.trace((key, field))

                    if field and key in field:
                        logger.debug("Removing key '{}'...", fieldname)
                        del field[key]
                        modified.add(fieldname)
            else:
                raise ValueError("INTERNAL ERROR: Unknown mode %r" % mode)

            # Commit any changes back into entry
            if modified:
                entry["torrent"].content = metainfo
                entry["torrent"].modified = True
                logger.info(
                    ('Key {} was' if len(modified) == 1 else 'Keys {} were')
                    + " scrubbed from torrent '{}'!",
                    ', '.join(sorted(modified)),
                    entry['title'],
                )
                new_infohash = entry["torrent"].info_hash
                if infohash != new_infohash:
                    logger.warning(
                        "Info hash changed from #{} to #{} in '{}'",
                        infohash,
                        new_infohash,
                        entry['filename'],
                    )


@event('plugin.register')
def register_plugin():
    plugin.register(TorrentScrub, interfaces=['task', 'torrent'], api_ver=2)



from argparse import ArgumentParser

from flexget import options
from flexget.db_schema import plugin_schemas, reset_schema
from flexget.event import event
from flexget.manager import Base, Session
from flexget.terminal import console


def do_cli(manager, options):
    with manager.acquire_lock():
        if options.db_action == 'cleanup':
            cleanup(manager)
        elif options.db_action == 'vacuum':
            vacuum()
        elif options.db_action == 'reset':
            reset(manager)
        elif options.db_action == 'reset-plugin':
            reset_plugin(options)


def cleanup(manager):
    manager.db_cleanup(force=True)
    console('Database cleanup complete.')


def vacuum():
    console('Running VACUUM on sqlite database, this could take a while.')
    session = Session()
    try:
        session.execute('VACUUM')
        session.commit()
    finally:
        session.close()
    console('VACUUM complete.')


def reset(manager):
    Base.metadata.drop_all(bind=manager.engine)
    Base.metadata.create_all(bind=manager.engine)
    console('The FlexGet database has been reset.')


def reset_plugin(options):
    plugin = options.plugin_name[0]
    try:
        reset_schema(plugin)
        console('The database for `%s` has been reset.' % plugin)
    except ValueError as e:
        console('Unable to reset %s: %s' % (plugin, e.message))


@event('options.register')
def register_parser_arguments():
    plugins_parser = ArgumentParser(add_help=False)
    plugins_parser.add_argument(
        'plugin_name', help="Name of plugin to reset", nargs=1, choices=list(plugin_schemas)
    )

    parser = options.register_command(
        'database', do_cli, help='Utilities to manage the FlexGet database'
    )
    subparsers = parser.add_subparsers(title='Actions', metavar='<action>', dest='db_action')
    subparsers.add_parser(
        'cleanup', help='Make all plugins clean un-needed data from the database'
    )
    subparsers.add_parser(
        'vacuum', help='Running vacuum can increase performance and decrease database size'
    )
    reset_parser = subparsers.add_parser(
        'reset', add_help=False, help='Reset the entire database (DANGEROUS!)'
    )
    reset_parser.add_argument(
        '--sure',
        action='store_true',
        required=True,
        help='You must use this flag to indicate you REALLY want to do this',
    )
    subparsers.add_parser(
        'reset-plugin', help='Reset the database for a specific plugin', parents=[plugins_parser]
    )



from flexget import options, plugin
from flexget.event import event
from flexget.manager import Session
from flexget.terminal import TerminalTable, console, table_parser

from . import db

try:
    # NOTE: Importing other plugins is discouraged!
    from flexget.components.rejected import remember_rejected as plugin_remember_rejected
except ImportError:
    raise plugin.DependencyError(issued_by=__name__, missing='remember_rejected')


def do_cli(manager, options):
    if options.rejected_action == 'list':
        list_rejected(options)
    elif options.rejected_action == 'clear':
        clear_rejected(manager)


def list_rejected(options):
    with Session() as session:
        results = session.query(db.RememberEntry).all()
        header = ['#', 'Title', 'Task', 'Rejected by', 'Reason']
        table = TerminalTable(*header, table_type=options.table_type)
        for entry in results:
            table.add_row(
                str(entry.id), entry.title, entry.task.name, entry.rejected_by, entry.reason or ''
            )
    console(table)


def clear_rejected(manager):
    with Session() as session:
        results = session.query(db.RememberEntry).delete()
        console('Cleared %i items.' % results)
        session.commit()
        if results:
            manager.config_changed()


@event('options.register')
def register_parser_arguments():
    parser = options.register_command(
        'rejected', do_cli, help='list or clear remembered rejections'
    )
    subparsers = parser.add_subparsers(dest='rejected_action', metavar='<action>')
    subparsers.add_parser(
        'list', help='list all the entries that have been rejected', parents=[table_parser]
    )
    subparsers.add_parser(
        'clear', help='clear all rejected entries from database, so they can be retried'
    )



from datetime import datetime

from loguru import logger
from sqlalchemy import Column, DateTime, Integer, String, Unicode
from sqlalchemy.schema import Index
from sqlalchemy.sql.expression import desc

from flexget import plugin
from flexget.event import event, fire_event
from flexget.manager import Base
from flexget.utils.log import log_once
from flexget.utils.tools import parse_timedelta

logger = logger.bind(name='proper_movies')


class ProperMovie(Base):
    __tablename__ = 'proper_movies'

    id = Column(Integer, primary_key=True)
    title = Column(Unicode)
    task = Column('feed', Unicode)
    imdb_id = Column(String, index=True)
    quality = Column(String)
    proper_count = Column(Integer)
    added = Column(DateTime)

    def __init__(self):
        self.added = datetime.now()

    def __repr__(self):
        return '<ProperMovie(title=%s,task=%s,imdb_id=%s,quality=%s,proper_count=%s,added=%s)>' % (
            self.title,
            self.task,
            self.imdb_id,
            self.quality,
            self.proper_count,
            self.added,
        )


# create index
columns = Base.metadata.tables['proper_movies'].c
Index(
    'proper_movies_imdb_id_quality_proper', columns.imdb_id, columns.quality, columns.proper_count
)


class FilterProperMovies:
    """
    Automatically download proper movies.

    Configuration:

        proper_movies: n <minutes|hours|days|weeks>

    or permanently:

        proper_movies: yes

    Value no will disable plugin.
    """

    schema = {'oneOf': [{'type': 'boolean'}, {'type': 'string', 'format': 'interval'}]}

    def on_task_filter(self, task, config):
        logger.debug('check for enforcing')

        # parse config
        if isinstance(config, bool):
            # configured a boolean false, disable plugin
            if not config:
                return
            # configured a boolean true, disable timeframe
            timeframe = None
        else:
            # parse time window
            logger.debug('interval: {}', config)
            try:
                timeframe = parse_timedelta(config)
            except ValueError:
                raise plugin.PluginError('Invalid time format', logger)

        # throws DependencyError if not present aborting task
        imdb_lookup = plugin.get_plugin_by_name('imdb_lookup').instance

        for entry in task.entries:
            parser = plugin.get('parsing', self).parse_movie(entry['title'])

            # if we have imdb_id already evaluated
            if entry.get('imdb_id', None, eval_lazy=False) is None:
                try:
                    # TODO: fix imdb_id_lookup, cumbersome that it returns None and or throws exception
                    # Also it's crappy name!
                    imdb_id = imdb_lookup.imdb_id_lookup(
                        movie_title=parser.name, movie_year=parser.year, raw_title=entry['title']
                    )
                    if imdb_id is None:
                        continue
                    entry['imdb_id'] = imdb_id
                except plugin.PluginError as pe:
                    log_once(pe.value)
                    continue

            quality = parser.quality.name

            logger.debug('quality: {}', quality)
            logger.debug('imdb_id: {}', entry['imdb_id'])
            logger.debug('current proper count: {}', parser.proper_count)

            proper_movie = (
                task.session.query(ProperMovie)
                .filter(ProperMovie.imdb_id == entry['imdb_id'])
                .filter(ProperMovie.quality == quality)
                .order_by(desc(ProperMovie.proper_count))
                .first()
            )

            if not proper_movie:
                logger.debug('no previous download recorded for {}', entry['imdb_id'])
                continue

            highest_proper_count = proper_movie.proper_count
            logger.debug('highest_proper_count: {}', highest_proper_count)

            accept_proper = False
            if parser.proper_count > highest_proper_count:
                logger.debug('proper detected: {} ', proper_movie)

                if timeframe is None:
                    accept_proper = True
                else:
                    expires = proper_movie.added + timeframe
                    logger.debug('propers timeframe: {}', timeframe)
                    logger.debug('added: {}', proper_movie.added)
                    logger.debug('propers ignore after: {}', str(expires))
                    if datetime.now() < expires:
                        accept_proper = True
                    else:
                        logger.verbose("Proper `{}` has past it's expiration time", entry['title'])

            if accept_proper:
                logger.info(
                    'Accepting proper version previously downloaded movie `{}`', entry['title']
                )
                # TODO: does this need to be called?
                # fire_event('forget', entry['imdb_url'])
                fire_event('forget', entry['imdb_id'])
                entry.accept('proper version of previously downloaded movie')

    def on_task_learn(self, task, config):
        """Add downloaded movies to the database"""
        logger.debug('check for learning')
        for entry in task.accepted:
            if 'imdb_id' not in entry:
                logger.debug('`{}` does not have imdb_id', entry['title'])
                continue

            parser = plugin.get('parsing', self).parse_movie(entry['title'])

            quality = parser.quality.name

            logger.debug('quality: {}', quality)
            logger.debug('imdb_id: {}', entry['imdb_id'])
            logger.debug('proper count: {}', parser.proper_count)

            proper_movie = (
                task.session.query(ProperMovie)
                .filter(ProperMovie.imdb_id == entry['imdb_id'])
                .filter(ProperMovie.quality == quality)
                .filter(ProperMovie.proper_count == parser.proper_count)
                .first()
            )

            if not proper_movie:
                pm = ProperMovie()
                pm.title = entry['title']
                pm.task = task.name
                pm.imdb_id = entry['imdb_id']
                pm.quality = quality
                pm.proper_count = parser.proper_count
                task.session.add(pm)
                logger.debug('added {}', pm)
            else:
                logger.debug('{} already exists', proper_movie)


@event('plugin.register')
def register_plugin():
    plugin.register(FilterProperMovies, 'proper_movies', api_ver=2)



import sys
from pathlib import Path
from typing import List

from setuptools import find_packages, setup

long_description = Path('README.rst').read_text()

# Populates __version__ without importing the package
__version__ = None
with open('flexget/_version.py', encoding='utf-8') as ver_file:
    exec(ver_file.read())  # pylint: disable=W0122
if not __version__:
    print('Could not find __version__ from flexget/_version.py')
    sys.exit(1)


def load_requirements(filename: str) -> List[str]:
    return [
        line.strip()
        for line in Path(filename).read_text().splitlines()
        if not line.startswith('#')
    ]


setup(
    name='FlexGet',
    version=__version__,
    description='FlexGet is a program aimed to automate downloading or processing content (torrents, podcasts, etc.) '
    'from different sources like RSS-feeds, html-pages, various sites and more.',
    long_description=long_description,
    long_description_content_type='text/x-rst',
    author='Marko Koivusalo',
    author_email='marko.koivusalo@gmail.com',
    license='MIT',
    url='https://flexget.com',
    project_urls={
        'Repository': 'https://github.com/Flexget/Flexget',
        'Issue Tracker': 'https://github.com/Flexget/Flexget/issues',
        'Forum': 'https://discuss.flexget.com',
    },
    packages=find_packages(exclude=['flexget.tests']),
    include_package_data=True,
    zip_safe=False,
    install_requires=load_requirements('requirements.txt'),
    tests_require=['pytest'],
    extras_require={'dev': load_requirements('dev-requirements.txt')},
    entry_points={
        'console_scripts': ['flexget = flexget:main'],
        'gui_scripts': [
            'flexget-headless = flexget:main'
        ],  # This is useful on Windows to avoid a cmd popup
    },
    python_requires='>=3.6',
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
    ],
)



import contextlib
import threading
from textwrap import wrap
from typing import Any, Iterator, Optional, TextIO

import rich
import rich.box
import rich.console
import rich.segment
import rich.table
import rich.text

from flexget.options import ArgumentParser

local_context = threading.local()
rich_console = rich.console.Console()

PORCELAIN_BOX: rich.box.Box = rich.box.Box(
    """\
    
  | 
    
  | 
    
    
  | 
    
""",
    ascii=True,
)

GITHUB_BOX: rich.box.Box = rich.box.Box(
    """\
    
| ||
|-||
| ||
|-||
|-||
| ||
    
""",
    ascii=True,
)


class TerminalTable(rich.table.Table):
    """
    A data table suited for CLI output, created via its sent parameters. For example::

        header = ['Col1', 'Col2']
        table_data = [header]
        for item in iterable:
            table_data.append([item.attribute1, item.attribute2])
        table = TerminalTable('plain', table_data)
        print table.output

    Optional values are setting table title, and supplying wrap_columns list and
    drop_column list. If table does not fit into terminal any columns listed in
    wrap_columns will be tried to wrap and if resulting columns are below MIN_WIDTH(10)
    columns listed in drop_column will be removed from output.

    Example::

        header = ['Col1', 'Col2']
        table_data = [header]
        for item in iterable:
            table_data.append([item.attribute1, item.attribute2])
        table = TerminalTable('plain', table_data, 'Table title', wrap_columns=[1,2],
                              drop_columns=[4,2])
        print table.output

    :param table_type: A string matching TABLE_TYPES keys.
    """

    # Easy access for our plugins without importing rich
    Column = rich.table.Column

    # TODO: Add other new types
    TABLE_TYPES = {
        'plain': {'box': rich.box.ASCII},
        'porcelain': {
            'box': PORCELAIN_BOX,
            'show_edge': False,
            'pad_edge': False,
            'title': None,
            'padding': 0,
        },
        'single': {'box': rich.box.SQUARE},
        'double': {'box': rich.box.DOUBLE},
        'github': {'box': GITHUB_BOX},
        'heavy-head': {'box': rich.box.HEAVY_HEAD},
    }

    def __init__(self, *args, table_type: str = None, **kwargs) -> None:
        self.table_type = table_type
        if table_type:
            kwargs = {**kwargs, **self.TABLE_TYPES[table_type]}
        super().__init__(*args, **kwargs)

    def __rich_console__(self, console, options):
        segments = super().__rich_console__(console, options)
        if self.table_type not in ['porcelain', 'github']:
            yield from segments
            return
        # Strips out blank lines from our custom types
        lines = rich.segment.Segment.split_lines(segments)
        for line in lines:
            if any(seg.text.strip() for seg in line):
                yield from line
                yield rich.segment.Segment.line()


table_parser = ArgumentParser(add_help=False)
table_parser.add_argument(
    '--table-type',
    choices=list(TerminalTable.TABLE_TYPES),
    default='heavy-head',
    help='Select output table style',
)
table_parser.add_argument(
    '--porcelain',
    dest='table_type',
    action='store_const',
    const='porcelain',
    help='Make the output parseable. Similar to using `--table-type porcelain`',
)


def word_wrap(text: str, max_length: int) -> str:
    """A helper method designed to return a wrapped string.

    :param text: Text to wrap
    :param max_length: Maximum allowed string length
    :return: Wrapped text or original text
    """
    if len(text) >= max_length:
        return '\n'.join(wrap(text, max_length))
    return text


def colorize(color: str, text: str) -> str:
    """
    A simple override of Color.colorize which sets the default auto colors value to True, since it's the more common
    use case. When output isn't TTY just return text

    :param color: Color tag to use
    :param text: Text to color
    :param auto: Whether to apply auto colors

    :return: Colored text or text
    """
    return f'[{color}]{text}[/]'


def disable_colors():
    """
    Disables colors to the terminal.
    """
    rich_console.no_color = True


@contextlib.contextmanager
def capture_console(filelike: TextIO) -> Iterator:
    old_output = get_console_output()
    local_context.output = filelike
    try:
        yield
    finally:
        local_context.output = old_output


def get_console_output() -> Optional[TextIO]:
    return getattr(local_context, 'output', None)


def _patchable_console(text, *args, **kwargs):
    # Nobody will import this directly, so we can monkeypatch it for IPC calls
    rich_console.file = get_console_output()
    try:
        rich_console.print(text, *args, **kwargs)
    finally:
        rich_console.file = None


def console(text: Any, *args, **kwargs) -> None:
    """
    Print to console safely. Output is able to be captured by different streams in different contexts.

    Any plugin wishing to output to the user's console should use this function instead of print so that
    output can be redirected when FlexGet is invoked from another process.

    Accepts arguments like the `rich.console.Console.print` function does.
    """
    _patchable_console(text, *args, **kwargs)



from loguru import logger

from flexget import entry, plugin
from flexget.event import event
from flexget.manager import Session
from flexget.utils.log import log_once
from flexget.utils.tools import split_title_year

logger = logger.bind(name='bluray_lookup')


class PluginBlurayLookup:
    """Retrieves bluray information for entries.

    Example:
        tmdb_lookup: yes
    """

    field_map = {
        'bluray_name': 'name',
        'bluray_id': 'id',
        'bluray_year': 'year',
        'bluray_disc_rating': 'bluray_rating',
        'bluray_rating': 'rating',
        'bluray_studio': 'studio',
        # Make sure we don't store the db association proxy directly on the entry
        'bluray_genres': lambda movie: list(movie.genres),
        'bluray_released': 'release_date',
        'bluray_certification': 'certification',
        'bluray_runtime': 'runtime',
        'bluray_url': 'url',
        # Generic fields filled by all movie lookup plugins:
        'movie_name': 'name',
        'movie_year': 'year',
    }

    schema = {'type': 'boolean'}

    @entry.register_lazy_lookup('bluray_lookup')
    def lazy_loader(self, entry):
        """Does the lookup for this entry and populates the entry fields."""
        lookup = plugin.get('api_bluray', self).lookup

        try:
            with Session() as session:
                title, year = split_title_year(entry['title'])
                movie = lookup(title=title, year=year, session=session)
                entry.update_using_map(self.field_map, movie)
        except LookupError:
            log_once('Bluray lookup failed for %s' % entry['title'], logger, 'WARNING')

    def lookup(self, entry):
        """
        Populates all lazy fields to an Entry. May be called by other plugins
        requiring bluray info on an Entry

        :param entry: Entry instance
        """
        entry.add_lazy_fields(self.lazy_loader, self.field_map)

    def on_task_metainfo(self, task, config):
        if not config:
            return
        for entry in task.entries:
            self.lookup(entry)

    @property
    def movie_identifier(self):
        """Returns the plugin main identifier type"""
        return 'bluray_id'


@event('plugin.register')
def register_plugin():
    plugin.register(
        PluginBlurayLookup, 'bluray_lookup', api_ver=2, interfaces=['task', 'movie_metainfo']
    )



import os
import xmlrpc.client
from socket import error as socket_error

from loguru import logger

from flexget import plugin
from flexget.event import event
from flexget.utils.template import RenderError

logger = logger.bind(name='aria2')


class OutputAria2:
    """
    Simple Aria2 output

    Example::

        aria2:
          path: ~/downloads/

    """

    schema = {
        'type': 'object',
        'properties': {
            'server': {'type': 'string', 'default': 'localhost'},
            'port': {'type': 'integer', 'default': 6800},
            'secret': {'type': 'string', 'default': ''},
            'username': {'type': 'string', 'default': ''},  # NOTE: To be deprecated by aria2
            'password': {'type': 'string', 'default': ''},
            'path': {'type': 'string'},
            'filename': {'type': 'string'},
            'options': {
                'type': 'object',
                'additionalProperties': {'oneOf': [{'type': 'string'}, {'type': 'integer'}]},
            },
        },
        'required': ['path'],
        'additionalProperties': False,
    }

    def aria2_connection(self, server, port, username=None, password=None):
        if username and password:
            userpass = '%s:%s@' % (username, password)
        else:
            userpass = ''
        url = 'http://%s%s:%s/rpc' % (userpass, server, port)
        logger.debug('aria2 url: {}', url)
        logger.info('Connecting to daemon at {}', url)
        try:
            return xmlrpc.client.ServerProxy(url).aria2
        except xmlrpc.client.ProtocolError as err:
            raise plugin.PluginError(
                'Could not connect to aria2 at %s. Protocol error %s: %s'
                % (url, err.errcode, err.errmsg),
                logger,
            )
        except xmlrpc.client.Fault as err:
            raise plugin.PluginError(
                'XML-RPC fault: Unable to connect to aria2 daemon at %s: %s'
                % (url, err.faultString),
                logger,
            )
        except socket_error as e:
            raise plugin.PluginError(
                'Socket connection issue with aria2 daemon at %s: %s' % (url, e), logger
            )
        except:
            logger.opt(exception=True).debug('Unexpected error during aria2 connection')
            raise plugin.PluginError(
                'Unidentified error during connection to aria2 daemon', logger
            )

    def prepare_config(self, config):
        config.setdefault('server', 'localhost')
        config.setdefault('port', 6800)
        config.setdefault('username', '')
        config.setdefault('password', '')
        config.setdefault('secret', '')
        config.setdefault('options', {})
        return config

    def on_task_output(self, task, config):
        # don't add when learning
        if task.options.learn:
            return
        config = self.prepare_config(config)
        aria2 = self.aria2_connection(
            config['server'], config['port'], config['username'], config['password']
        )
        for entry in task.accepted:
            if task.options.test:
                logger.verbose('Would add `{}` to aria2.', entry['title'])
                continue
            try:
                self.add_entry(aria2, entry, config)
            except socket_error as se:
                entry.fail('Unable to reach Aria2: %s' % se)
            except xmlrpc.client.Fault as err:
                logger.critical('Fault code {} message {}', err.faultCode, err.faultString)
                entry.fail('Aria2 communication Fault')
            except Exception as e:
                logger.opt(exception=True).debug('Exception type {}', type(e))
                raise

    def add_entry(self, aria2, entry, config):
        """
        Add entry to Aria2
        """
        options = config['options']
        try:
            options['dir'] = os.path.expanduser(entry.render(config['path']).rstrip('/'))
        except RenderError as e:
            entry.fail('failed to render \'path\': %s' % e)
            return
        if 'filename' in config:
            try:
                options['out'] = os.path.expanduser(entry.render(config['filename']))
            except RenderError as e:
                entry.fail('failed to render \'filename\': %s' % e)
                return
        secret = None
        if config['secret']:
            secret = 'token:%s' % config['secret']
        # handle torrent files
        if 'torrent' in entry:
            if 'file' in entry:
                torrent_file = entry['file']
            elif 'location' in entry:
                # in case download plugin moved the file elsewhere
                torrent_file = entry['location']
            else:
                entry.fail('Cannot find torrent file')
                return
            if secret:
                return aria2.addTorrent(
                    secret, xmlrpc.client.Binary(open(torrent_file, mode='rb').read()), [], options
                )
            return aria2.addTorrent(
                xmlrpc.client.Binary(open(torrent_file, mode='rb').read()), [], options
            )
        # handle everything else (except metalink -- which is unsupported)
        # so magnets, https, http, ftp .. etc
        if secret:
            return aria2.addUri(secret, [entry['url']], options)
        return aria2.addUri([entry['url']], options)


@event('plugin.register')
def register_plugin():
    plugin.register(OutputAria2, 'aria2', api_ver=2)



import socket

from loguru import logger
from requests.exceptions import RequestException

from flexget import plugin
from flexget.event import event
from flexget.plugin import PluginWarning
from flexget.utils.requests import Session as RequestSession

requests = RequestSession(max_retries=3)

plugin_name = "cronitor"

logger = logger.bind(name=plugin_name)


class Cronitor:
    """
    Example::
      cronitor: ABC123

    Or:
      cronitor:
        monitor_code: ABC123
        on_start: yes
        on_abort: no
        message: Ping
        host: foo.bar
        auth_key: secret
    """

    base_url = "https://cronitor.link/{monitor_code}/{status}"

    schema = {
        "oneOf": [
            {
                "type": "object",
                "properties": {
                    "monitor_code": {"type": "string"},
                    "on_start": {"type": "boolean"},
                    "on_abort": {"type": "boolean"},
                    "message": {"type": "string"},
                    "host": {"type": "string"},
                    "auth_key": {"type": "string"},
                },
                "required": ["monitor_code"],
                "additionalProperties": False,
            },
            {"type": "string"},
        ]
    }

    @staticmethod
    def prepare_config(config):
        if isinstance(config, str):
            config = {"monitor_code": config}
        config.setdefault("on_start", True)
        config.setdefault("on_abort", True)
        config.setdefault("host", socket.gethostname())
        return config

    def _send_request(self, status, config, task_name):
        url = self.base_url.format(monitor_code=config["monitor_code"], status=status)
        message = config.get(
            "message", "{task} task {status}".format(task=task_name, status=status)
        )
        data = {"msg": message, "host": config["host"]}
        if config.get("auth_key"):
            data["auth_key"] = config["auth_key"]
        try:
            rsp = requests.get(url, params=data)
            rsp.raise_for_status()
        except RequestException as e:
            raise PluginWarning("Could not report to cronitor: {}".format(e))

    def on_task_start(self, task, config):
        config = self.prepare_config(config)
        if not config["on_start"]:
            return
        self._send_request("run", config, task.name)

    def on_task_abort(self, task, config):
        config = self.prepare_config(config)
        if not config["on_abort"]:
            return
        self._send_request("fail", config, task.name)

    def on_task_exit(self, task, config):
        config = self.prepare_config(config)
        self._send_request("complete", config, task.name)


@event("plugin.register")
def register_plugin():
    plugin.register(Cronitor, plugin_name, api_ver=2)



from flask import Response
from sqlalchemy.orm import Session

from flexget.api import APIResource, api
from flexget.api.app import base_message_schema, success_response

schema_api = api.namespace(
    'format_check', description='Test Flexget custom schema format validations'
)


class ObjectContainer:
    format_checker_input = {
        'type': 'object',
        'properties': {
            'quality': {'type': 'string', 'format': 'quality'},
            'quality_requirements': {'type': 'string', 'format': 'quality_requirements'},
            'time': {'type': 'string', 'format': 'time'},
            'interval': {'type': 'string', 'format': 'interval'},
            'size': {'type': 'string', 'format': 'size'},
            'percent': {'type': 'string', 'format': 'percent'},
            'regex': {'type': 'string', 'format': 'regex'},
            'file': {'type': 'string', 'format': 'file'},
            'path': {'type': 'string', 'format': 'path'},
            'url': {'type': 'string', 'format': 'url'},
            'episode_identifier': {'type': 'string', 'format': 'episode_identifier'},
            'episode_or_season_id': {'type': 'string', 'format': 'episode_or_season_id'},
        },
    }


format_checker_schema = api.schema_model('format_checker', ObjectContainer.format_checker_input)


@schema_api.route('/', doc=False)
class SchemaTest(APIResource):
    @api.validate(format_checker_schema)
    @api.response(200, model=base_message_schema)
    def post(self, session: Session = None) -> Response:
        """Validate flexget custom schema"""
        # If validation passed, all is well
        return success_response('payload is valid')



from loguru import logger

from flexget.api import api_app
from flexget.config_schema import register_config_key
from flexget.event import event
from flexget.ui.v1 import register_web_ui as register_web_ui_v1
from flexget.ui.v2 import register_web_ui as register_web_ui_v2
from flexget.utils.tools import get_config_hash
from flexget.webserver import get_secret, register_app, setup_server

logger = logger.bind(name="web_server_daemon")
config_hash = ''
web_server = None

web_config_schema = {
    'oneOf': [
        {'type': 'boolean'},
        {'type': 'integer', 'minimum': 0, 'maximum': 65536},
        {
            'type': 'object',
            'properties': {
                'bind': {'type': 'string', 'format': 'ipv4'},
                'port': {'type': 'integer', 'minimum': 0, 'maximum': 65536},
                'ssl_certificate': {'type': 'string'},
                'ssl_private_key': {'type': 'string'},
                'web_ui': {'type': 'boolean'},
                'base_url': {'type': 'string'},
                'run_v2': {
                    'type': 'boolean',
                    'deprecated': 'v2 is registered by default if web_ui: true so `run_v2` is now redundant. To run v1 alongside, use the `run_v1`.',
                },
                'run_v1': {'type': 'boolean'},
            },
            'additionalProperties': False,
            'dependencies': {
                'ssl_certificate': ['ssl_private_key'],
                'ssl_private_key': ['ssl_certificate'],
            },
        },
    ]
}


def prepare_config(config):
    if not config:
        return
    if isinstance(config, bool):
        config = {}
    if isinstance(config, int):
        config = {'port': config}
    config.setdefault('bind', '0.0.0.0')
    config.setdefault('port', 5050)
    config.setdefault('ssl_certificate', None)
    config.setdefault('ssl_private_key', None)
    config.setdefault('web_ui', True)
    config.setdefault('base_url', '')
    config.setdefault('run_v2', False)
    config.setdefault('run_v1', False)
    if config['base_url']:
        if not config['base_url'].startswith('/'):
            config['base_url'] = '/' + config['base_url']
        if config['base_url'].endswith('/'):
            config['base_url'] = config['base_url'][:-1]

    return config


@event('config.register')
def register_config():
    register_config_key('web_server', web_config_schema)


@event('manager.config_updated')
@event('manager.daemon.started')
def register_web_server(manager):
    """Registers Web Server and loads API (always) and WebUi via config"""
    global web_server, config_hash

    if not manager.is_daemon:
        return

    config = manager.config.get('web_server')
    if get_config_hash(config) == config_hash:
        logger.debug('web server config has\'nt changed')
        return

    config_hash = get_config_hash(config)
    web_server_config = prepare_config(config)

    # Removes any existing web server instances if exists
    stop_server(manager)

    if not web_server_config:
        return

    logger.info(
        'Running web server at IP {}:{}', web_server_config['bind'], web_server_config['port']
    )
    # Register API
    api_app.secret_key = get_secret()

    logger.info("Initiating API")
    register_app('/api', api_app, 'API')

    # Register WebUI
    if web_server_config.get('web_ui'):
        if web_server_config.get('run_v1'):
            logger.info('Registering WebUI v1')
            register_web_ui_v1(manager)

        logger.info('Registering WebUI v2')
        register_web_ui_v2(web_server_config)

    web_server = setup_server(web_server_config)


@event('manager.shutdown')
def stop_server(manager):
    """Sets up and starts/restarts the webui."""
    global web_server

    if not manager.is_daemon:
        return

    if web_server and web_server.is_alive():
        web_server.stop()
    web_server = None



import re

from loguru import logger

from flexget import plugin
from flexget.event import event

logger = logger.bind(name='abort_if_exists')


class PluginAbortIfExists:
    """Aborts a task if an entry field matches the regexp"""

    schema = {
        'type': 'object',
        'properties': {
            'regexp': {'type': 'string', 'format': 'regex'},
            'field': {'type': 'string'},
        },
        'required': ['regexp', 'field'],
        'additionalProperties': False,
    }

    # Execute as the first thing in filter phase
    @plugin.priority(plugin.PRIORITY_FIRST)
    def on_task_filter(self, task, config):
        abort_re = re.compile(config['regexp'], re.IGNORECASE)
        field = config['field']
        for entry in task.all_entries:
            # if pattern matches any entry
            if field not in entry:
                logger.debug('Field {} not found. Skipping.', field)
                continue
            if abort_re.search(entry[field]):
                task.abort('An entry contained %s in field %s. Abort!' % (config['regexp'], field))


@event('plugin.register')
def register_plugin():
    plugin.register(PluginAbortIfExists, 'abort_if_exists', api_ver=2)



import os
import re

from loguru import logger
from sqlalchemy import Column, Integer, Unicode

from flexget import options, plugin
from flexget.db_schema import versioned_base
from flexget.entry import Entry
from flexget.event import event
from flexget.manager import Session

logger = logger.bind(name='tail')
Base = versioned_base('tail', 0)


class TailPosition(Base):
    __tablename__ = 'tail'
    id = Column(Integer, primary_key=True)
    task = Column(Unicode)
    filename = Column(Unicode)
    position = Column(Integer)


class InputTail:
    """
    Parse any text for entries using regular expression.

    ::

      file: <file>
      entry:
        <field>: <regexp to match value>
      format:
        <field>: <python string formatting>

    Note: each entry must have at least two fields, title and url

    You may wish to specify encoding used by file so file can be properly
    decoded. List of encodings
    at http://docs.python.org/library/codecs.html#standard-encodings.

    Example::

      tail:
        file: ~/irclogs/some/log
        entry:
          title: 'TITLE: (.*) URL:'
          url: 'URL: (.*)'
        encoding: utf8
    """

    schema = {
        'type': 'object',
        'properties': {
            'file': {'type': 'string', 'format': 'file'},
            'encoding': {'type': 'string'},
            'entry': {
                'type': 'object',
                'properties': {
                    'url': {'type': 'string', 'format': 'regex'},
                    'title': {'type': 'string', 'format': 'regex'},
                },
                'required': ['url', 'title'],
            },
            'format': {'type': 'object', 'additionalProperties': {'type': 'string'}},
        },
        'required': ['file', 'entry'],
        'additionalProperties': False,
    }

    def format_entry(self, entry, d):
        for k, v in d.items():
            entry[k] = v % entry

    def on_task_input(self, task, config):

        # Let details plugin know that it is ok if this task doesn't produce any entries
        task.no_entries_ok = True

        filename = os.path.expanduser(config['file'])
        encoding = config.get('encoding', 'utf-8')
        with Session() as session:
            db_pos = (
                session.query(TailPosition)
                .filter(TailPosition.task == task.name)
                .filter(TailPosition.filename == filename)
                .first()
            )
            if db_pos:
                last_pos = db_pos.position
            else:
                last_pos = 0

            with open(filename, 'r', encoding=encoding, errors='replace') as file:
                if task.options.tail_reset == filename or task.options.tail_reset == task.name:
                    if last_pos == 0:
                        logger.info('Task {} tail position is already zero', task.name)
                    else:
                        logger.info(
                            'Task {} tail position ({}) reset to zero', task.name, last_pos
                        )
                        last_pos = 0

                if os.path.getsize(filename) < last_pos:
                    logger.info(
                        'File size is smaller than in previous execution, resetting to beginning of the file'
                    )
                    last_pos = 0

                file.seek(last_pos)

                logger.debug('continuing from last position {}', last_pos)

                entry_config = config.get('entry')
                format_config = config.get('format', {})

                # keep track what fields have been found
                used = {}
                entries = []
                entry = Entry()

                # now parse text

                for line in file:
                    if not line:
                        break

                    for field, regexp in entry_config.items():
                        # log.debug('search field: %s regexp: %s' % (field, regexp))
                        match = re.search(regexp, line)
                        if match:
                            # check if used field detected, in such case start with new entry
                            if field in used:
                                if entry.isvalid():
                                    logger.info(
                                        'Found field {} again before entry was completed. Adding current incomplete, but valid entry and moving to next.',
                                        field,
                                    )
                                    self.format_entry(entry, format_config)
                                    entries.append(entry)
                                else:
                                    logger.info(
                                        'Invalid data, entry field {} is already found once. Ignoring entry.',
                                        field,
                                    )
                                # start new entry
                                entry = Entry()
                                used = {}

                            # add field to entry
                            entry[field] = match.group(1)
                            used[field] = True
                            logger.debug('found field: {} value: {}', field, entry[field])

                        # if all fields have been found
                        if len(used) == len(entry_config):
                            # check that entry has at least title and url
                            if not entry.isvalid():
                                logger.info(
                                    'Invalid data, constructed entry is missing mandatory fields (title or url)'
                                )
                            else:
                                self.format_entry(entry, format_config)
                                entries.append(entry)
                                logger.debug('Added entry {}', entry)
                                # start new entry
                                entry = Entry()
                                used = {}
                last_pos = file.tell()
            if db_pos:
                db_pos.position = last_pos
            else:
                session.add(TailPosition(task=task.name, filename=filename, position=last_pos))
        return entries


@event('plugin.register')
def register_plugin():
    plugin.register(InputTail, 'tail', api_ver=2)


@event('options.register')
def register_parser_arguments():
    options.get_parser('execute').add_argument(
        '--tail-reset',
        action='store',
        dest='tail_reset',
        default=False,
        metavar='FILE|TASK',
        help='reset tail position for a file',
    )



from datetime import datetime, timedelta

from loguru import logger
from sqlalchemy import Column, DateTime, ForeignKey, Index, Integer, String, Unicode
from sqlalchemy.orm import relation

from flexget import db_schema
from flexget.event import event
from flexget.utils.sqlalchemy_utils import table_add_column, table_columns

logger = logger.bind(name='remember_rej')
Base = db_schema.versioned_base('remember_rejected', 3)


@db_schema.upgrade('remember_rejected')
def upgrade(ver, session):
    if ver is None:
        columns = table_columns('remember_rejected_entry', session)
        if 'uid' in columns:
            raise db_schema.UpgradeImpossible
        ver = 0
    if ver == 0:
        logger.info('Adding reason column to remember_rejected_entry table.')
        table_add_column('remember_rejected_entry', 'reason', String, session)
        ver = 1
    if ver == 1:
        logger.info('Adding `added` column to remember_rejected_entry table.')
        table_add_column(
            'remember_rejected_entry', 'added', DateTime, session, default=datetime.now
        )
        ver = 2
    if ver == 2:
        logger.info('Adding expires column to remember_rejected_entry table.')
        table_add_column('remember_rejected_entry', 'expires', DateTime, session)
        ver = 3
    return ver


class RememberTask(Base):
    __tablename__ = 'remember_rejected_feeds'

    id = Column(Integer, primary_key=True)
    name = Column(Unicode)

    entries = relation('RememberEntry', backref='task', cascade='all, delete, delete-orphan')


class RememberEntry(Base):
    __tablename__ = 'remember_rejected_entry'

    id = Column(Integer, primary_key=True)
    added = Column(DateTime, default=datetime.now)
    expires = Column(DateTime)
    title = Column(Unicode)
    url = Column(String)
    rejected_by = Column(String)
    reason = Column(String)

    task_id = Column('feed_id', Integer, ForeignKey('remember_rejected_feeds.id'), nullable=False)


Index('remember_feed_title_url', RememberEntry.task_id, RememberEntry.title, RememberEntry.url)


@event('manager.db_cleanup')
def db_cleanup(manager, session):
    # Remove entries older than 30 days
    result = (
        session.query(RememberEntry)
        .filter(RememberEntry.added < datetime.now() - timedelta(days=30))
        .delete()
    )
    if result:
        logger.verbose('Removed {} entries from remember rejected table.', result)


def get_rejected(session, count=None, start=None, stop=None, sort_by=None, descending=None):
    query = session.query(RememberEntry)
    if count:
        return query.count()
    if descending:
        query = query.order_by(getattr(RememberEntry, sort_by).desc())
    else:
        query = query.order_by(getattr(RememberEntry, sort_by))
    return query.slice(start, stop).all()



import re

from loguru import logger

from flexget import plugin
from flexget.components.sites.urlrewriting import UrlRewritingError
from flexget.event import event

logger = logger.bind(name='urlrewrite')


class UrlRewrite:
    r"""
    Generic configurable urlrewriter.

    Example::

      urlrewrite:
        demonoid:
          regexp: http://www\.demonoid\.com/files/details/
          format: http://www.demonoid.com/files/download/HTTP/
    """

    resolves = {}

    schema = {
        'type': 'object',
        'additionalProperties': {
            'type': 'object',
            'properties': {
                'regexp': {'type': 'string', 'format': 'regex'},
                'format': {'type': 'string'},
            },
            'required': ['regexp', 'format'],
            'additionalProperties': False,
        },
    }

    def on_task_start(self, task, config):
        resolves = self.resolves[task.name] = {}
        for name, rewrite_config in config.items():
            match = re.compile(rewrite_config['regexp'])
            format = rewrite_config['format']
            resolves[name] = {
                'regexp_compiled': match,
                'format': format,
                'regexp': rewrite_config['regexp'],
            }
            logger.debug('Added rewrite {}', name)

    def url_rewritable(self, task, entry):
        logger.trace('running url_rewritable')
        logger.trace(self.resolves)
        for _, config in self.resolves.get(task.name, {}).items():
            regexp = config['regexp_compiled']
            logger.trace('testing {}', config['regexp'])
            if regexp.search(entry['url']):
                return True
        return False

    def url_rewrite(self, task, entry):
        for name, config in self.resolves.get(task.name, {}).items():
            regexp = config['regexp_compiled']
            format = config['format']
            if regexp.search(entry['url']):
                logger.debug('Regexp resolving {} with {}', entry['url'], name)

                # run the regexp
                entry['url'] = regexp.sub(format, entry['url'])

                if regexp.match(entry['url']):
                    entry.fail('urlrewriting')
                    raise UrlRewritingError(
                        'Regexp %s result should NOT continue to match!' % name
                    )
                return


@event('plugin.register')
def register_plugin():
    plugin.register(UrlRewrite, 'urlrewrite', interfaces=['task', 'urlrewriter'], api_ver=2)



import sys

from loguru import logger

from flexget import plugin
from flexget.event import event

logger = logger.bind(name='change')
found_deprecated = False


class ChangeWarn:
    """
    Gives warning if user has deprecated / changed configuration in the root level.

    Will be replaced by root level validation in the future!

    Contains ugly hacks, better to include all deprecation warnings here during 1.0 BETA phase
    """

    def on_task_start(self, task, config):
        global found_deprecated

        if 'torrent_size' in task.config:
            logger.critical('Plugin torrent_size is deprecated, use content_size instead')
            found_deprecated = True

        if 'nzb_size' in task.config:
            logger.critical('Plugin nzb_size is deprecated, use content_size instead')
            found_deprecated = True

        if found_deprecated:
            task.manager.shutdown(finish_queue=False)
            task.abort('Deprecated config.')


@event('plugin.register')
def register_plugin():
    plugin.register(ChangeWarn, 'change_warn', builtin=True, api_ver=2)


# check that no old plugins are in pre-compiled form (pyc)
try:
    import os.path

    plugin_dirs = (
        os.path.normpath(sys.path[0] + '/../flexget/plugins/'),
        os.path.normpath(sys.path[0] + '/../flexget/plugins/input/'),
    )
    for plugin_dir in plugin_dirs:
        for name in os.listdir(plugin_dir):
            require_clean = False

            if name.startswith('module'):
                require_clean = True

            if name == 'csv.pyc':
                require_clean = True

            if 'resolver' in name:
                require_clean = True

            if 'filter_torrent_size' in name:
                require_clean = True

            if 'filter_nzb_size' in name:
                require_clean = True

            if 'module_priority' in name:
                require_clean = True

            if 'ignore_feed' in name:
                require_clean = True

            if 'module_manual' in name:
                require_clean = True

            if 'output_exec' in name:
                require_clean = True

            if 'plugin_adv_exec' in name:
                require_clean = True

            if 'output_transmissionrpc' in name:
                require_clean = True

            if require_clean:
                logger.critical('-' * 79)
                logger.critical('IMPORTANT: Your installation has some files from older FlexGet!')
                logger.critical('')
                logger.critical(
                    ' Please remove all pre-compiled .pyc and .pyo files from {}', plugin_dir
                )
                logger.critical(' Offending file: {}', name)
                logger.critical('')
                logger.critical(
                    '           After getting rid of these FlexGet should run again normally'
                )

                logger.critical('')
                logger.critical('-' * 79)
                found_deprecated = True
                break

except Exception:
    pass



from loguru import logger

from flexget import plugin
from flexget.event import event

from . import db

logger = logger.bind(name='history')


class PluginHistory:
    """Records all accepted entries for later lookup"""

    schema = {'type': 'boolean'}

    def on_task_learn(self, task, config):
        """Add accepted entries to history"""
        if config is False:
            return  # Explicitly disabled with configuration

        for entry in task.accepted:
            item = db.History()
            item.task = task.name
            item.filename = entry.get('output', None)
            item.title = entry['title']
            item.url = entry['url']
            reason = ''
            if 'reason' in entry:
                reason = ' (reason: %s)' % entry['reason']
            item.details = 'Accepted by %s%s' % (entry.get('accepted_by', '<unknown>'), reason)
            task.session.add(item)


@event('plugin.register')
def register_plugin():
    plugin.register(PluginHistory, 'history', builtin=True, api_ver=2)



from loguru import logger

from flexget import plugin
from flexget.event import event

logger = logger.bind(name='mock_output')


class MockOutput:
    """
    Debugging plugin which records a copy of all accepted entries into a list stored in `mock_output` attribute
    of the task.
    """

    schema = {'type': 'boolean'}

    def on_task_start(self, task, config):
        task.mock_output = []

    def on_task_output(self, task, config):
        task.mock_output.extend(e.copy() for e in task.all_entries if e.accepted)

    def on_task_exit(self, task, config):
        logger.verbose(
            'The following titles were output during this task run: {}',
            ', '.join(e['title'] for e in task.mock_output),
        )


@event('plugin.register')
def register_plugin():
    plugin.register(MockOutput, 'mock_output', debug=True, api_ver=2)



from loguru import logger

from flexget import plugin
from flexget.components.sites.utils import normalize_scene
from flexget.config_schema import one_or_more
from flexget.entry import Entry
from flexget.event import event
from flexget.plugin import PluginError
from flexget.utils.requests import RequestException, Session, TimedLimiter

logger = logger.bind(name='rarbg')

requests = Session()
requests.add_domain_limiter(
    TimedLimiter('torrentapi.org', '3 seconds')
)  # they only allow 1 request per 2 seconds

CATEGORIES = {
    'all': 0,
    # Movies
    'x264': 17,
    'x264 720p': 45,
    'x264 1080p': 44,
    'x264 3D': 47,
    'XviD': 14,
    'XviD 720p': 48,
    'Full BD': 42,
    # TV
    'HDTV': 41,
    'SDTV': 18,
    # Adult
    'XXX': 4,
    # Music
    'MusicMP3': 23,
    'MusicFLAC': 25,
    # Games
    'Games/PC ISO': 27,
    'Games/PC RIP': 28,
    'Games/PS3': 40,
    'Games/XBOX-360': 32,
    'Software/PC ISO': 33,
    # E-Books
    'e-Books': 35,
}


class SearchRarBG:
    """
    RarBG search plugin. Implements https://torrentapi.org/apidocs_v2.txt

    To perform search against single category:

    rarbg:
        category: x264 720p

    To perform search against multiple categories:

    rarbg:
        category:
            - x264 720p
            - x264 1080p

    Movie categories accepted: x264 720p, x264 1080p, XviD, Full BD
    TV categories accepted: HDTV, SDTV

    You can use also use category ID manually if you so desire (eg. x264 720p is actually category id '45')
    """

    schema = {
        'type': 'object',
        'properties': {
            'category': one_or_more(
                {'oneOf': [{'type': 'integer'}, {'type': 'string', 'enum': list(CATEGORIES)}]}
            ),
            'sorted_by': {
                'type': 'string',
                'enum': ['seeders', 'leechers', 'last'],
                'default': 'last',
            },
            # min_seeders and min_leechers seem to be working again
            'min_seeders': {'type': 'integer', 'default': 0},
            'min_leechers': {'type': 'integer', 'default': 0},
            'limit': {'type': 'integer', 'enum': [25, 50, 100], 'default': 25},
            'ranked': {'type': 'boolean', 'default': True},
            'use_tvdb': {'type': 'boolean', 'default': False},
        },
        "additionalProperties": False,
    }

    base_url = 'https://torrentapi.org/pubapi_v2.php'
    token = None

    def get_token(self, refresh=False):
        if refresh or not self.token:
            try:
                response = requests.get(
                    self.base_url,
                    params={'get_token': 'get_token', 'format': 'json', 'app_id': 'flexget'},
                ).json()
                self.token = response.get('token')
                logger.debug('RarBG token: {}', self.token)
            except RequestException as e:
                logger.opt(exception=True).debug('Could not retrieve RarBG token')
                raise PluginError('Could not retrieve token: %s' % e)
        return self.token

    def get(self, params, token_error=False):
        """
        Simple get-wrapper that allows updating invalid tokens

        :param params: the params to be passed to requests
        :param token_error: whether or not we previously have had token errors, if True we should fetch a new one
        :return: json response
        """
        params['token'] = self.get_token(refresh=token_error)
        try:
            response = requests.get(self.base_url, params=params)
            logger.debug('requesting: {}', response.url)
            response = response.json()
        except RequestException as e:
            logger.error('Rarbg request failed: {}', e)
            return

        # error code 1, 2 and 4 pertain to token errors
        if response.get('error_code') in [1, 2, 4]:
            logger.debug(
                'Invalid token. Error {}: {}', response['error_code'], response.get('error')
            )
            if token_error:
                raise PluginError('Could not retrieve a valid token: %s' % response.get('error'))
            return self.get(params=params, token_error=True)

        return response

    @plugin.internet(logger)
    def search(self, task, entry, config):
        """
        Search for entries on RarBG
        """

        categories = config.get('category', 'all')
        # Ensure categories a list
        if not isinstance(categories, list):
            categories = [categories]
        # Convert named category to its respective category id number
        categories = [c if isinstance(c, int) else CATEGORIES[c] for c in categories]
        category_url_fragment = ';'.join(str(c) for c in categories)

        entries = set()

        params = {
            'mode': 'search',
            'ranked': int(config['ranked']),
            'min_seeders': config['min_seeders'],
            'min_leechers': config['min_leechers'],
            'sort': config['sorted_by'],
            'category': category_url_fragment,
            'format': 'json_extended',
            'app_id': 'flexget',
        }

        for search_string in entry.get('search_strings', [entry['title']]):
            params.pop('search_string', None)
            params.pop('search_imdb', None)
            params.pop('search_tvdb', None)

            if entry.get('movie_name') and entry.get('imdb_id'):
                params['search_imdb'] = entry.get('imdb_id')
            else:
                query = normalize_scene(search_string)
                query_url_fragment = query.encode('utf8')
                params['search_string'] = query_url_fragment
                if config['use_tvdb']:
                    plugin.get('thetvdb_lookup', self).lazy_series_lookup(entry, 'en')
                    params['search_tvdb'] = entry.get('tvdb_id')
                    logger.debug('Using tvdb id {}', entry.get('tvdb_id'))

            response = self.get(params=params)
            if not response:
                continue

            # error code 10 and 20 just mean no results were found
            if response.get('error_code') in [10, 20]:
                searched_string = (
                    params.get('search_string')
                    or 'imdb={0}'.format(params.get('search_imdb'))
                    or 'tvdb={0}'.format(params.get('tvdb_id'))
                )
                logger.debug(
                    'No results found for {}. Message from rarbg: {}',
                    searched_string,
                    response.get('error'),
                )
                continue
            elif response.get('error'):
                logger.error(
                    'Error code {}: {}', response.get('error_code'), response.get('error')
                )
                continue
            else:
                for result in response.get('torrent_results'):
                    e = Entry()

                    e['title'] = result.get('title')
                    e['url'] = result.get('download')
                    e['torrent_seeds'] = int(result.get('seeders'))
                    e['torrent_leeches'] = int(result.get('leechers'))
                    e['content_size'] = int(result.get('size')) / 1024 / 1024
                    episode_info = result.get('episode_info')
                    if episode_info:
                        e['imdb_id'] = episode_info.get('imdb')
                        e['tvdb_id'] = episode_info.get('tvdb')
                        e['tvrage_id'] = episode_info.get('tvrage')

                    entries.add(e)

        return entries


@event('plugin.register')
def register_plugin():
    plugin.register(SearchRarBG, 'rarbg', interfaces=['search'], api_ver=2)



from loguru import logger

from flexget import plugin
from flexget.event import event
from flexget.task import Task

logger = logger.bind(name='max_reruns')


class MaxReRuns:
    """Overrides the maximum amount of re-runs allowed by a task."""

    schema = {'type': 'integer'}

    def __init__(self):
        self.default = Task.RERUN_DEFAULT

    def reset(self, task):
        task.unlock_reruns()
        task.max_reruns = self.default
        logger.debug('changing max task rerun variable back to: {}', self.default)

    def on_task_start(self, task, config):
        self.default = task.max_reruns
        logger.debug('saving old max task rerun value: {}', self.default)
        task.max_reruns = int(config)
        task.lock_reruns()
        logger.debug('changing max task rerun variable to: {}', config)

    def on_task_exit(self, task, config):
        if task.rerun_count > task.max_reruns:
            self.reset(task)

    def on_task_abort(self, task, config):
        self.reset(task)


@event('plugin.register')
def register_plugin():
    plugin.register(MaxReRuns, 'max_reruns', api_ver=2)



import copy

from flexget.api.app import base_message
from flexget.components.seen.api import ObjectsContainer as OC
from flexget.components.seen.db import SeenEntry, SeenField
from flexget.manager import Session
from flexget.utils import json


class TestSeenAPI:
    config = "{'tasks': {}}"

    def add_seen_entries(self):
        seen_entry_1 = dict(title='test_title', reason='test_reason', task='test_task')
        field_1 = dict(field='test_field_1', value='test_value_1')
        field_2 = dict(field='test_field_2', value='test_value_2')
        seen_entry_2 = dict(
            title='test_title_2', reason='test_reason_2', task='test_task_2', local=True
        )
        field_3 = dict(field='test_field_3', value='test_value_3')
        field_4 = dict(field='test_field_4', value='test_value_4')

        entries = sorted([seen_entry_1, seen_entry_2], key=lambda entry: entry['title'])

        with Session() as session:
            seen_db_1 = SeenEntry(**seen_entry_1)
            seen_db_1.fields = [SeenField(**field_1), SeenField(**field_2)]
            session.add(seen_db_1)
            seen_db_2 = SeenEntry(**seen_entry_2)
            seen_db_2.fields = [SeenField(**field_3), SeenField(**field_4)]
            session.add(seen_db_2)

        return entries

    def test_seen_get_all(self, api_client, schema_match):
        rsp = api_client.get('/seen/')
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(OC.seen_search_object, data)
        assert not errors

        entries = self.add_seen_entries()

        rsp = api_client.get('/seen/')
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(OC.seen_search_object, data)
        assert not errors

        for idx, value in enumerate(sorted(data, key=lambda entry: entry['title'])):
            for k, v in entries[idx].items():
                assert value[k] == v

        assert len(data) == 2

        rsp = api_client.get('/seen/?local=true')
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(OC.seen_search_object, data)
        assert not errors

        assert len(data) == 1

        rsp = api_client.get('/seen/?value=test_value_2')
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(OC.seen_search_object, data)
        assert not errors

        assert len(data) == 1

        rsp = api_client.get('/seen/?value=bla')
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(OC.seen_search_object, data)
        assert not errors

        assert len(data) == 0

    def test_seen_delete_all(self, api_client, schema_match):
        entries = self.add_seen_entries()

        rsp = api_client.get('/seen/')
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(OC.seen_search_object, data)
        assert not errors

        for idx, value in enumerate(sorted(data, key=lambda entry: entry['title'])):
            for k, v in entries[idx].items():
                assert value[k] == v

        rsp = api_client.delete('/seen/')
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(base_message, data)
        assert not errors

        assert data['message'] == 'successfully deleted 2 entries'

        self.add_seen_entries()

        rsp = api_client.delete('/seen/?local=true')
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(base_message, data)
        assert not errors

        assert data['message'] == 'successfully deleted 1 entries'

        rsp = api_client.delete('/seen/?value=test_value_2')
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(base_message, data)
        assert not errors

        assert data['message'] == 'successfully deleted 1 entries'

        rsp = api_client.get('/seen/')
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(OC.seen_search_object, data)
        assert not errors

        assert data == []

    def test_seen_get_by_id(self, api_client, schema_match):
        entries = self.add_seen_entries()

        rsp = api_client.get('/seen/1/')
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(OC.seen_object, data)
        assert not errors

        for key, value in entries[0].items():
            assert data[key] == value

        rsp = api_client.get('/seen/10/')
        assert rsp.status_code == 404, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(base_message, data)
        assert not errors

    def test_seen_get_delete_id(self, api_client, schema_match):
        self.add_seen_entries()

        rsp = api_client.delete('/seen/1/')
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(base_message, data)
        assert not errors

        rsp = api_client.get('/seen/')
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(OC.seen_search_object, data)
        assert not errors

        assert len(data) == 1

        rsp = api_client.delete('/seen/10/')
        assert rsp.status_code == 404, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(base_message, data)
        assert not errors


class TestSeenPagination:
    config = 'tasks: {}'

    def test_seen_pagination(self, api_client, link_headers):
        base_seen_entry = dict(title='test_title_', task='test_task_', reason='test_reason_')
        base_seen_field = dict(field='test_field_', value='test_value_')
        number_of_entries = 200

        with Session() as session:
            for i in range(number_of_entries):
                entry = copy.deepcopy(base_seen_entry)
                field = copy.deepcopy(base_seen_field)

                for key, value in entry.items():
                    entry[key] = value + str(i)

                for key, value in field.items():
                    field[key] = value + str(i)

                seen_entry = SeenEntry(**entry)
                session.add(seen_entry)
                seen_entry.fields = [SeenField(**field)]

        # Default values
        rsp = api_client.get('/seen/')
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        assert len(data) == 50
        assert int(rsp.headers['total-count']) == 200
        assert int(rsp.headers['count']) == 50

        links = link_headers(rsp)
        assert links['last']['page'] == 4
        assert links['next']['page'] == 2

        # Change page size
        rsp = api_client.get('/seen/?per_page=100')
        assert rsp.status_code == 200
        data = json.loads(rsp.get_data(as_text=True))

        assert len(data) == 100
        assert int(rsp.headers['total-count']) == 200
        assert int(rsp.headers['count']) == 100

        links = link_headers(rsp)
        assert links['last']['page'] == 2
        assert links['next']['page'] == 2

        # Get different page
        rsp = api_client.get('/seen/?page=2')
        assert rsp.status_code == 200
        data = json.loads(rsp.get_data(as_text=True))

        assert len(data) == 50
        assert int(rsp.headers['total-count']) == 200
        assert int(rsp.headers['count']) == 50

        links = link_headers(rsp)
        assert links['last']['page'] == 4
        assert links['next']['page'] == 3
        assert links['prev']['page'] == 1

    def test_seen_sorting(self, api_client):
        seen_entry_1 = dict(
            title='test_title_1', reason='test_reason_c', task='test_task_2', local=True
        )
        field_1 = dict(field='test_field_1', value='test_value_1')
        field_2 = dict(field='test_field_2', value='test_value_2')
        seen_entry_2 = dict(
            title='test_title_2', reason='test_reason_b', task='test_task_3', local=True
        )
        field_3 = dict(field='test_field_3', value='test_value_3')
        field_4 = dict(field='test_field_4', value='test_value_4')
        seen_entry_3 = dict(
            title='test_title_3', reason='test_reason_a', task='test_task_1', local=False
        )
        field_5 = dict(field='test_field_3', value='test_value_3')
        field_6 = dict(field='test_field_4', value='test_value_4')

        with Session() as session:
            seen_db_1 = SeenEntry(**seen_entry_1)
            session.add(seen_db_1)
            seen_db_1.fields = [SeenField(**field_1), SeenField(**field_2)]

            seen_db_2 = SeenEntry(**seen_entry_2)
            session.add(seen_db_2)
            seen_db_2.fields = [SeenField(**field_3), SeenField(**field_4)]

            seen_db_3 = SeenEntry(**seen_entry_3)
            session.add(seen_db_3)
            seen_db_3.fields = [SeenField(**field_5), SeenField(**field_6)]

        # Sort by title
        rsp = api_client.get('/seen/?sort_by=title')
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        assert data[0]['title'] == 'test_title_3'

        rsp = api_client.get('/seen/?sort_by=title&order=asc')
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        assert data[0]['title'] == 'test_title_1'

        # Sort by task
        rsp = api_client.get('/seen/?sort_by=task')
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        assert data[0]['task'] == 'test_task_3'

        rsp = api_client.get('/seen/?sort_by=task&order=asc')
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        assert data[0]['task'] == 'test_task_1'

        # Sort by reason
        rsp = api_client.get('/seen/?sort_by=reason')
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        assert data[0]['reason'] == 'test_reason_c'

        rsp = api_client.get('/seen/?sort_by=reason&order=asc')
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        assert data[0]['reason'] == 'test_reason_a'

        # Sort by local
        rsp = api_client.get('/seen/?sort_by=local')
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        assert data[0]['local'] == True

        rsp = api_client.get('/seen/?sort_by=local&order=asc')
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        assert data[0]['local'] == False

        # Combine sorting and pagination
        rsp = api_client.get('/seen/?sort_by=reason&per_page=2&page=2')
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        assert data[0]['reason'] == 'test_reason_a'



import os

from loguru import logger

from flexget import plugin
from flexget.event import event
from flexget.utils.bittorrent import Torrent, is_torrent_file

logger = logger.bind(name='modif_torrent')


class TorrentFilename:
    """
    Makes sure that entries containing torrent-file have .torrent
    extension. This is enabled always by default (builtins).
    """

    TORRENT_PRIO = 255

    @plugin.priority(TORRENT_PRIO)
    def on_task_modify(self, task, config):
        # Only scan through accepted entries, as the file must have been downloaded in order to parse anything
        for entry in task.accepted:
            # skip if entry does not have file assigned
            if 'file' not in entry:
                logger.trace("{} doesn't have a file associated", entry['title'])
                continue
            if not os.path.exists(entry['file']):
                logger.debug('File {} does not exist', entry['file'])
                continue
            if os.path.getsize(entry['file']) == 0:
                logger.debug('File {} is 0 bytes in size', entry['file'])
                continue
            if not is_torrent_file(entry['file']):
                continue
            logger.debug('{} seems to be a torrent', entry['title'])

            # create torrent object from torrent
            try:
                with open(entry['file'], 'rb') as f:
                    # NOTE: this reads entire file into memory, but we're pretty sure it's
                    # a small torrent file since it starts with TORRENT_RE
                    data = f.read()

                if 'content-length' in entry:
                    if len(data) != entry['content-length']:
                        entry.fail(
                            'Torrent file length doesn\'t match to the one reported by the server'
                        )
                        self.purge(entry)
                        continue

                # construct torrent object
                try:
                    torrent = Torrent(data)
                except SyntaxError as e:
                    entry.fail('%s - broken or invalid torrent file received' % e.args[0])
                    self.purge(entry)
                    continue

                logger.trace('working on torrent {}', torrent)
                entry['torrent'] = torrent
                entry['torrent_info_hash'] = torrent.info_hash
                # if we do not have good filename (by download plugin)
                # for this entry, try to generate one from torrent content
                if entry.get('filename'):
                    if not entry['filename'].lower().endswith('.torrent'):
                        # filename present but without .torrent extension, add it
                        entry['filename'] += '.torrent'
                else:
                    # generate filename from torrent or fall back to title plus extension
                    entry['filename'] = self.make_filename(torrent, entry)
            except Exception as e:
                logger.exception(e)

    @plugin.priority(TORRENT_PRIO)
    def on_task_output(self, task, config):
        for entry in task.entries:
            if 'torrent' in entry:
                if entry['torrent'].modified:
                    # re-write data into a file
                    logger.debug('Writing modified torrent file for {}', entry['title'])
                    with open(entry['file'], 'wb+') as f:
                        f.write(entry['torrent'].encode())

    def make_filename(self, torrent, entry):
        """Build a filename for this torrent"""

        title = entry['title']
        files = torrent.get_filelist()
        if len(files) == 1:
            # single file, if filename is longer than title use it
            fn = files[0]['name']
            if len(fn) > len(title):
                title = fn[: fn.rfind('.')]

        # neatify title
        title = title.replace('/', '_')
        title = title.replace(' ', '_')
        title = title.replace('\u200b', '')

        # title = title.encode('iso8859-1', 'ignore') # Damn \u200b -character, how I loathe thee
        # TODO: replace only zero width spaces, leave unicode alone?

        fn = '%s.torrent' % title
        logger.debug('make_filename made {}', fn)
        return fn

    def purge(self, entry):
        if os.path.exists(entry['file']):
            logger.debug('removing temp file {} from {}', entry['file'], entry['title'])
            os.remove(entry['file'])
        del entry['file']


@event('plugin.register')
def register_plugin():
    plugin.register(TorrentFilename, 'torrent', builtin=True, api_ver=2)



from flexget.api.app import base_message
from flexget.api.core.user import ObjectsContainer as OC
from flexget.utils import json


class TestUserAPI:
    config = 'tasks: {}'

    def test_change_password(self, execute_task, api_client, schema_match):
        weak_password = {'password': 'weak'}
        medium_password = {'password': 'a.better.password'}
        strong_password = {'password': 'AVer123y$ron__g-=PaW[]rd'}

        rsp = api_client.json_put('/user/', data=json.dumps(weak_password))
        assert rsp.status_code == 400
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(base_message, data)
        assert not errors

        rsp = api_client.json_put('/user/', data=json.dumps(medium_password))
        assert rsp.status_code == 200
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(base_message, data)
        assert not errors

        rsp = api_client.json_put('/user/', data=json.dumps(strong_password))
        assert rsp.status_code == 200
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(base_message, data)
        assert not errors

    def test_change_token(self, execute_task, api_client, schema_match):
        rsp = api_client.json_put('user/token/')
        assert rsp.status_code == 200
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(OC.user_token_response, data)
        assert not errors



from loguru import logger

from flexget import plugin
from flexget.event import event

from . import db

logger = logger.bind(name='series_remove')


class OutputSeriesRemove:
    schema = {
        'oneOf': [
            {'type': 'boolean'},
            {'type': 'object', 'properties': {'forget': {'type': 'boolean'}}},
        ],
    }

    def on_task_output(self, task, config):
        if not config:
            return
        if isinstance(config, bool):
            config = {'forget': False}
        forget = config['forget']

        for entry in task.accepted:
            if 'series_name' in entry:
                if 'series_id' in entry:
                    try:
                        db.remove_series_entity(entry['series_name'], entry['series_id'], forget)
                        logger.info(
                            'Removed {}episode `{}` from series `{}` download history.',
                            'and forgot ' if forget else '',
                            entry['series_id'],
                            entry['series_name'],
                        )
                    except ValueError:
                        logger.debug(
                            'Series ({}) or id ({}) unknown.',
                            entry['series_name'],
                            entry['series_id'],
                        )
                else:
                    try:
                        db.remove_series(entry['series_name'], forget)
                        logger.info(
                            'Removed {}series `{}` download history.',
                            'and forgot ' if forget else '',
                            entry['series_name'],
                        )
                    except ValueError:
                        logger.debug('Series ({}) unknown.', entry['series_name'])


@event('plugin.register')
def register_plugin():
    plugin.register(OutputSeriesRemove, 'series_remove', api_ver=2)



from bs4 import BeautifulSoup
from loguru import logger

from flexget import plugin
from flexget.event import event
from flexget.utils import requests

logger = logger.bind(name='nnm-club')


class UrlRewriteNnmClub:
    """Nnm-club.me urlrewriter."""

    def url_rewritable(self, task, entry):
        return entry['url'].startswith('http://nnm-club.me/forum/viewtopic.php?t=')

    def url_rewrite(self, task, entry):
        try:
            r = task.requests.get(entry['url'])
        except requests.RequestException as e:
            logger.error('Error while fetching page: {}', e)
            entry['url'] = None
            return
        html = r.content
        soup = BeautifulSoup(html)
        links = soup.findAll('a', href=True)
        magnets = [x for x in links if x.get('href').startswith('magnet')]
        if not magnets:
            logger.error('There is no magnet links on page ({})', entry['url'])
            entry['url'] = None
            return
        entry['url'] = magnets[0]


@event('plugin.register')
def register_plugin():
    plugin.register(UrlRewriteNnmClub, 'nnm-club', interfaces=['urlrewriter'], api_ver=2)



import time
from argparse import SUPPRESS

from loguru import logger
from sqlalchemy.engine import Connection

from flexget import options
from flexget.event import add_event_handler, event, remove_event_handler

logger = logger.bind(name='performance')

performance = {}

_start = {}

query_count = 0
orig_execute = None


def log_query_count(name_point):
    """Debugging purposes, allows logging number of executed queries at :name_point:"""
    logger.info('At point named `{}` total of {} queries were ran', name_point, query_count)


def before_plugin(task, keyword):
    fd = _start.setdefault(task.name, {})
    fd.setdefault('time', {})[keyword] = time.time()
    fd.setdefault('queries', {})[keyword] = query_count


def after_plugin(task, keyword):
    took = time.time() - _start[task.name]['time'][keyword]
    queries = query_count - _start[task.name]['queries'][keyword]
    # Store results, increases previous values
    pd = performance.setdefault(task.name, {})
    data = pd.setdefault(keyword, {})
    data['took'] = data.get('took', 0) + took
    data['queries'] = data.get('queries', 0) + queries


@event('manager.execute.started')
def startup(manager, options):
    if not options.debug_perf:
        return

    logger.info('Enabling plugin and SQLAlchemy performance debugging')
    global query_count, orig_execute
    query_count = 0

    # Monkeypatch query counter for SQLAlchemy
    if hasattr(Connection, 'execute'):
        orig_execute = Connection.execute

        def monkeypatched(*args, **kwargs):
            global query_count
            query_count += 1
            return orig_execute(*args, **kwargs)

        Connection.execute = monkeypatched
    else:
        logger.critical('Unable to monkeypatch sqlalchemy')

    add_event_handler('task.execute.before_plugin', before_plugin)
    add_event_handler('task.execute.after_plugin', after_plugin)


@event('manager.execute.completed')
def cleanup(manager, options):
    if not options.debug_perf:
        return

    # Print summary
    for name, data in performance.items():
        logger.info('Performance results for task {}:', name)
        for keyword, results in data.items():
            took = results['took']
            queries = results['queries']
            if took > 0.1 or queries > 10:
                logger.info('{:<15} took {:0.2f} sec ({} queries)', keyword, took, queries)

    # Deregister our hooks
    if hasattr(Connection, 'execute') and orig_execute:
        Connection.execute = orig_execute
    remove_event_handler('task.execute.before_plugin', before_plugin)
    remove_event_handler('task.execute.after_plugin', after_plugin)


@event('options.register')
def register_parser_arguments():
    options.get_parser('execute').add_argument(
        '--debug-perf', action='store_true', dest='debug_perf', default=False, help=SUPPRESS
    )



"""
Helper module that can load whatever version of the json module is available.
Plugins can just import the methods from this module.

Also allows date and datetime objects to be encoded/decoded.
"""
import datetime
from collections.abc import Iterable, Mapping
from contextlib import suppress
from typing import Any, Union

from flexget.plugin import DependencyError

try:
    import simplejson as json
except ImportError:
    try:
        import json  # type: ignore  # python/mypy#1153
    except ImportError:
        try:
            # Google Appengine offers simplejson via django
            from django.utils import simplejson as json  # type: ignore
        except ImportError:
            raise DependencyError(missing='simplejson')

DATE_FMT = '%Y-%m-%d'
ISO8601_FMT = '%Y-%m-%dT%H:%M:%SZ'


class DTDecoder(json.JSONDecoder):
    def decode(self, obj, **kwargs):
        # The built-in `json` library will `unicode` strings, except for empty strings. patch this for
        # consistency so that `unicode` is always returned.
        if obj == b'':
            return ''

        if isinstance(obj, str):
            dt_str = obj.strip('"')
            try:
                return datetime.datetime.strptime(dt_str, ISO8601_FMT)
            except (ValueError, TypeError):
                with suppress(ValueError, TypeError):
                    return datetime.datetime.strptime(dt_str, DATE_FMT)

        return super().decode(obj, **kwargs)


def _datetime_encoder(obj: Union[datetime.datetime, datetime.date]) -> str:
    if isinstance(obj, datetime.datetime):
        return obj.strftime(ISO8601_FMT)
    elif isinstance(obj, datetime.date):
        return obj.strftime(DATE_FMT)
    raise TypeError


def _datetime_decoder(dict_: dict) -> dict:
    for key, value in dict_.items():
        # The built-in `json` library will `unicode` strings, except for empty strings. patch this for
        # consistency so that `unicode` is always returned.
        if value == b'':
            dict_[key] = ''
            continue

        try:
            datetime_obj = datetime.datetime.strptime(value, ISO8601_FMT)
            dict_[key] = datetime_obj
        except (ValueError, TypeError):
            with suppress(ValueError, TypeError):
                date_obj = datetime.datetime.strptime(value, DATE_FMT)
                dict_[key] = date_obj.date()

    return dict_


def _empty_unicode_decoder(dict_: dict) -> dict:
    for key, value in dict_.items():
        # The built-in `json` library will `unicode` strings, except for empty strings. patch this for
        # consistency so that `unicode` is always returned.
        if value == b'':
            dict_[key] = ''
    return dict_


def dumps(*args, **kwargs) -> str:
    if kwargs.pop('encode_datetime', False):
        kwargs['default'] = _datetime_encoder
    return json.dumps(*args, **kwargs)


def dump(*args, **kwargs) -> None:
    if kwargs.pop('encode_datetime', False):
        kwargs['default'] = _datetime_encoder
    return json.dump(*args, **kwargs)


def loads(*args, **kwargs) -> Any:
    """
    :param bool decode_datetime: If `True`, dates in ISO8601 format will be deserialized to :class:`datetime.datetime`
      objects.
    """
    if kwargs.pop('decode_datetime', False):
        kwargs['object_hook'] = _datetime_decoder
        kwargs['cls'] = DTDecoder
    else:
        kwargs['object_hook'] = _empty_unicode_decoder
    return json.loads(*args, **kwargs)


def load(*args, **kwargs) -> Any:
    """
    :param bool decode_datetime: If `True`, dates in ISO8601 format will be deserialized to :class:`datetime.datetime`
      objects.
    """
    if kwargs.pop('decode_datetime', False):
        kwargs['object_hook'] = _datetime_decoder
        kwargs['cls'] = DTDecoder
    else:
        kwargs['object_hook'] = _empty_unicode_decoder
    return json.load(*args, **kwargs)


def coerce(obj) -> Union[str, int, float, bool, dict, list, None]:
    """
    Coerce a data structure to a JSON serializable form.

    Will recursively go through data structure, and attempt to turn anything not JSON serializable into a string.
    """
    if isinstance(obj, (str, int, float, bool, type(None))):
        return obj
    if isinstance(obj, datetime.datetime):
        return obj.strftime(ISO8601_FMT)
    if isinstance(obj, datetime.date):
        return obj.strftime(DATE_FMT)
    if isinstance(obj, Mapping):
        return {k: coerce(v) for k, v in obj.items()}
    if isinstance(obj, Iterable):
        return [coerce(v) for v in obj]
    try:
        return str(obj)
    except Exception:
        return 'NOT JSON SERIALIZABLE'



from loguru import logger

from flexget import plugin
from flexget.entry import Entry
from flexget.event import event
from flexget.utils.cached_input import cached
from flexget.utils.requests import RequestException, Session, TimedLimiter
from flexget.utils.soup import get_soup

logger = logger.bind(name='letterboxd')

requests = Session(max_retries=5)
requests.add_domain_limiter(TimedLimiter('letterboxd.com', '1 seconds'))
base_url = 'http://letterboxd.com'

SLUGS = {
    'default': {'p_slug': '/%(user)s/list/%(list)s/', 'f_slug': 'data-film-slug'},
    'diary': {'p_slug': '/%(user)s/films/diary/', 'f_slug': 'data-film-slug'},
    'likes': {'p_slug': '/%(user)s/likes/films/', 'f_slug': 'data-film-link'},
    'rated': {'p_slug': '/%(user)s/films/ratings/', 'f_slug': 'data-film-slug'},
    'watched': {'p_slug': '/%(user)s/films/', 'f_slug': 'data-film-slug'},
    'watchlist': {'p_slug': '/%(user)s/watchlist/', 'f_slug': 'data-film-slug'},
}

SORT_BY = {
    'default': '',
    'added': 'by/added/',
    'length-ascending': 'by/shortest/',
    'length-descending': 'by/longest/',
    'name': 'by/name/',
    'popularity': 'by/popular/',
    'rating-ascending': 'by/rating-lowest/',
    'rating-descending': 'by/rating/',
    'release-ascending': 'by/release-earliest/',
    'release-descending': 'by/release/',
}


class Letterboxd:
    schema = {
        'type': 'object',
        'properties': {
            'username': {'type': 'string'},
            'list': {'type': 'string'},
            'sort_by': {'type': 'string', 'enum': list(SORT_BY.keys()), 'default': 'default'},
            'max_results': {
                'type': 'integer',
                'deprecated': '`limit` plugin should be used instead of letterboxd `max_results` option',
            },
        },
        'required': ['username', 'list'],
        'additionalProperties': False,
    }

    def build_config(self, config):
        config['list'] = config['list'].lower().replace(' ', '-')
        list_key = config['list']
        if list_key not in list(SLUGS.keys()):
            list_key = 'default'
        config['p_slug'] = SLUGS[list_key]['p_slug'] % {
            'user': config['username'],
            'list': config['list'],
        }
        config['f_slug'] = SLUGS[list_key]['f_slug']
        config['sort_by'] = SORT_BY[config['sort_by']]

        return config

    def tmdb_lookup(self, search):
        if not search:
            logger.warning('Can\'t search tmdb, no tmdb_id')
            return

        try:
            tmdb = plugin.get('api_tmdb', self).lookup(tmdb_id=search)
        except LookupError as e:
            logger.warning('Error searching tmdb: {}', e)
            return

        result = {
            'title': '%s (%s)' % (tmdb.name, tmdb.year),
            'imdb_id': tmdb.imdb_id,
            'tmdb_id': tmdb.id,
            'movie_name': tmdb.name,
            'movie_year': tmdb.year,
        }

        return result

    def parse_film(self, film, config):
        url = base_url + film.get(config['f_slug'])
        soup = get_soup(requests.get(url).content)
        result = self.tmdb_lookup(soup.find(attrs={'data-tmdb-id': True}).get('data-tmdb-id'))
        if not result:
            return

        entry = Entry(result)
        entry['url'] = url
        entry['letterboxd_list'] = '%s (%s)' % (config['list'], config['username'])
        try:
            entry['letterboxd_score'] = float(soup.find(itemprop='average').get('content'))
        except AttributeError:
            pass
        if config['list'] == 'diary':
            entry['letterboxd_uscore'] = int(
                film.find_next(attrs={'data-rating': True}).get('data-rating')
            )
        elif config['list'] == 'rated':
            entry['letterboxd_uscore'] = int(film.find_next(itemprop='rating').get('content'))

        return entry

    @cached('letterboxd', persist='2 hours')
    def on_task_input(self, task, config=None):
        config = self.build_config(config)
        url = base_url + config['p_slug'] + config['sort_by']
        max_results = config.get('max_results', 1)
        rcount = 0
        next_page = ''

        logger.verbose('Looking for films in Letterboxd list: {}', url)

        while next_page is not None and rcount < max_results:
            try:
                page = requests.get(url).content
            except RequestException as e:
                raise plugin.PluginError('Error retrieving list from Letterboxd: %s' % e)
            soup = get_soup(page)

            for film in soup.find_all(attrs={config['f_slug']: True}):
                if rcount < max_results:
                    result = self.parse_film(film, config)
                    if not result:
                        continue
                    yield result
                    if 'max_results' in config:
                        rcount += 1

            next_page = soup.select_one('.paginate-nextprev .next')
            if next_page is not None:
                next_page = next_page.get('href')
                if next_page is not None:
                    url = base_url + next_page


@event('plugin.register')
def register_plugin():
    plugin.register(Letterboxd, 'letterboxd', api_ver=2)



import re

from loguru import logger

from flexget import plugin
from flexget.components.imdb.utils import extract_id
from flexget.components.sites.utils import torrent_availability
from flexget.entry import Entry
from flexget.event import event
from flexget.utils import requests
from flexget.utils.soup import get_soup
from flexget.utils.tools import parse_filesize

logger = logger.bind(name='search_ptn')

session = requests.Session()

categories = {
    '1080p': 'c5',
    '720p': 'c6',
    'bdrip': 'c10',
    'bluray': 'c1',
    'brrip': 'c11',
    'dvdr': 'c4',
    'dvdrip': 'c12',
    'mp4': 'c16',
    'ost/flac': 'c17',
    'ost/mp3': 'c18',
    'packs': 'c20',
    'r5/scr': 'c13',
    'remux': 'c2',
    'tvrip': 'c15',
    'webrip': 'c14',
}

default_search_params = {
    # 'searchstring': 'search term',
    # 'advancedsearchparameters': '[year=1999]',
    'sort': 'browsedate',
    'skw': 'showall',
    'compression': 'unraredonly',
    'packs': 'torrentsonly',
    'titleonly': 'true',
    'subscriptions': 'showall',
    'visibility': 'aliveonly',
    'visiblecategories': 'Action,Adventure,Animation,Biography,Comedy,Crime,Documentary,Drama,Eastern,Family,Fantasy,'
    'History,Holiday,Horror,Kids,Musical,Mystery,Romance,Sci-Fi,Short,Sports,Thriller,War,Western',
    'order': 'DESC',
    'action': 'torrentstable',
    'bookmarks': 'showall',
    'viewtype': 1,
    'page': 1,
}


class SearchPTN:
    schema = {
        'type': 'object',
        'properties': {
            'username': {'type': 'string'},
            'login_key': {'type': 'string'},
            'password': {'type': 'string'},
            'categories': {
                'type': 'array',
                'items': {'type': 'string', 'enum': list(categories)},
                'deprecated': 'PtN category filtering is broken. Someone open a PR!',
            },
        },
        'required': ['username', 'login_key', 'password'],
        'additionalProperties': False,
    }

    def create_entries(self, soup, imdb_id=None):
        entries = []
        links = soup.findAll('a', attrs={'href': re.compile(r'download\.php\?torrent=\d+')})
        rows = [l.find_parent('tr') for l in links]
        for row in rows:
            entry = Entry()
            entry['title'] = row.find('a', attrs={'href': re.compile(r'detail\.php\?id')}).text
            dl_href = row.find('a', attrs={'href': re.compile(r'download\.php\?torrent=\d+')}).get(
                'href'
            )
            entry['url'] = 'http://piratethenet.org' + dl_href
            entry['torrent_seeds'] = int(row.find(title='Number of Seeders').text)
            entry['torrent_leeches'] = int(row.find(title='Number of Leechers').text)
            entry['torrent_availability'] = torrent_availability(
                entry['torrent_seeds'], entry['torrent_leeches']
            )

            entry['content_size'] = parse_filesize(
                str(row.find(title='Torrent size').text), si=False
            )

            if imdb_id:
                entry['imdb_id'] = imdb_id
            entries.append(entry)
        return entries

    def search(self, task, entry, config):
        if not session.cookies:
            try:
                login_params = {
                    'username': config['username'],
                    'password': config['password'],
                    'loginkey': config['login_key'],
                }
                r = session.post(
                    'https://piratethenet.org/takelogin.php', data=login_params, verify=False
                )
            except requests.RequestException as e:
                logger.error('Error while logging in to PtN: {}', e)
                raise plugin.PluginError('Could not log in to PtN')

            passkey = re.search(r'passkey=([\d\w]+)"', r.text)
            if not passkey:
                logger.error("It doesn't look like PtN login worked properly.")
                raise plugin.PluginError('PTN cookie info invalid')

        search_params = default_search_params.copy()
        if 'movie_name' in entry:
            if 'movie_year' in entry:
                search_params['advancedsearchparameters'] = '[year=%s]' % entry['movie_year']
            searches = [entry['movie_name']]
        else:
            searches = entry.get('search_strings', [entry['title']])

        results = set()
        for search in searches:
            search_params['searchstring'] = search
            try:
                r = session.get('http://piratethenet.org/torrentsutils.php', params=search_params)
            except requests.RequestException as e:
                logger.error('Error searching ptn: {}', e)
                continue
            # html5parser doesn't work properly for some reason
            soup = get_soup(r.text, parser='html.parser')
            for movie in soup.select('.torrentstd'):
                imdb_id = movie.find('a', href=re.compile(r'.*imdb\.com/title/tt'))
                if imdb_id:
                    imdb_id = extract_id(imdb_id['href'])
                if imdb_id and 'imdb_id' in entry and imdb_id != entry['imdb_id']:
                    continue
                results.update(self.create_entries(movie, imdb_id=imdb_id))

        return results


@event('plugin.register')
def register_plugin():
    plugin.register(SearchPTN, 'ptn', interfaces=['search'], api_ver=2)



import argparse
import os
import socket
import sys

from loguru import logger

import flexget
from flexget import options
from flexget.event import event
from flexget.terminal import console

logger = logger.bind(name='win32_service')

try:
    import servicemanager
    import win32event
    import win32service
    import win32serviceutil

    class AppServerSvc(win32serviceutil.ServiceFramework):
        _svc_name_ = 'FlexGet'
        _svc_display_name_ = 'FlexGet Daemon'
        _svc_description_ = 'Runs FlexGet tasks according to defined schedules'

        def __init__(self, args):
            win32serviceutil.ServiceFramework.__init__(self, args)
            self.hWaitStop = win32event.CreateEvent(None, 0, 0, None)
            socket.setdefaulttimeout(60)
            self.manager = None

        def SvcStop(self):
            self.ReportServiceStatus(win32service.SERVICE_STOP_PENDING)
            from flexget.manager import manager

            manager.shutdown(finish_queue=False)
            self.ReportServiceStatus(win32service.SERVICE_STOPPED)

        def SvcDoRun(self):
            servicemanager.LogMsg(
                servicemanager.EVENTLOG_INFORMATION_TYPE,
                servicemanager.PYS_SERVICE_STARTED,
                (self._svc_name_, ''),
            )

            flexget.main(['daemon', 'start'])


except ImportError:
    pass


def do_cli(manager, options):
    import win32file
    import win32serviceutil

    if hasattr(sys, 'real_prefix'):
        # We are in a virtualenv, there is some special setup
        if not os.path.exists(os.path.join(sys.prefix, 'python.exe')):
            console('Creating a hard link to virtualenv python.exe in root of virtualenv')
            win32file.CreateHardLink(
                os.path.join(sys.prefix, 'python.exe'),
                os.path.join(sys.prefix, 'Scripts', 'python.exe'),
            )

    argv = options.args
    if options.help:
        argv = []

    # Hack sys.argv a bit so that we get a better usage message
    sys.argv[0] = 'flexget service'
    win32serviceutil.HandleCommandLine(AppServerSvc, argv=['flexget service'] + argv)


@event('options.register')
def register_parser_arguments():
    if not sys.platform.startswith('win'):
        return
    # Still not fully working. Hidden for now.
    parser = options.register_command(
        'service',
        do_cli,  # help='set up or control a windows service for the daemon',
        add_help=False,
    )
    parser.add_argument('--help', '-h', action='store_true')
    parser.add_argument('args', nargs=argparse.REMAINDER)



import pytest
import requests


@pytest.mark.online
class TestCachedAPI:
    config = 'tasks: {}'

    def test_cached_api(self, api_client):
        rsp = api_client.get('/cached/')
        assert rsp.status_code == 400, 'Response code is %s' % rsp.status_code

        rsp = api_client.get('/cached/?url=bla')
        assert rsp.status_code == 400, 'Response code is %s' % rsp.status_code

        image_url = 'http://thetvdb.com/banners/fanart/original/281662-18.jpg'
        response = requests.get(image_url)

        assert response.status_code == 200

        rsp = api_client.get('/cached/?url={}'.format(image_url))
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code

        assert rsp.data == response.content

        rsp = api_client.get('/cached/?url={}?force=true'.format(image_url))
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code



import json

from flexget.api.app import base_message
from flexget.api.core.tasks import ObjectsContainer as OC


class TestExecuteAPI:
    @staticmethod
    def get_task_queue(manager):
        """Used to execute task queue"""
        assert len(manager.task_queue) == 1
        task = manager.task_queue.run_queue.get(timeout=0.5)
        assert task
        return task

    config = """
        tasks:
          test_task:
            mock:
              - title: accept_me
              - title: reject_me
            regexp:
              accept:
                - accept
              reject:
                - reject
        """

    def test_execute(self, api_client, manager, schema_match):
        # Minimal payload
        payload = {'tasks': ['test_task']}
        rsp = api_client.json_post('/tasks/execute/', data=json.dumps(payload))
        assert rsp.status_code == 200
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(OC.task_execution_results_schema, data)
        assert not errors

        task = self.get_task_queue(manager)
        task.execute()

        assert len(task.accepted) == 1

    def test_inject_plain(self, api_client, manager, schema_match):
        entry = {'title': "injected", 'url': 'http://test.com'}

        payload = {"inject": [entry], 'tasks': ['test_task']}
        rsp = api_client.json_post('/tasks/execute/', data=json.dumps(payload))
        assert rsp.status_code == 200
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(OC.task_execution_results_schema, data)
        assert not errors

        task = self.get_task_queue(manager)
        task.execute()

        assert len(task.all_entries) == 1
        assert len(task.accepted) == 0

    def test_inject_accept(self, api_client, manager, schema_match):
        entry = {
            'title': "injected",
            'url': 'http://test.com',
            'accept': True,
            'tasks': ['test_task'],
        }

        payload = {"inject": [entry], 'tasks': ['test_task']}
        rsp = api_client.json_post('/tasks/execute/', data=json.dumps(payload))
        assert rsp.status_code == 200
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(OC.task_execution_results_schema, data)
        assert not errors

        task = self.get_task_queue(manager)
        task.execute()

        assert len(task.all_entries) == 1
        assert len(task.accepted) == 1

    def test_inject_force(self, api_client, manager, schema_match):
        entry = {'title': "accept", 'url': 'http://test.com'}

        payload = {"inject": [entry], 'tasks': ['test_task']}
        rsp = api_client.json_post('/tasks/execute/', data=json.dumps(payload))
        assert rsp.status_code == 200
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(OC.task_execution_results_schema, data)
        assert not errors

        task = self.get_task_queue(manager)
        task.execute()

        assert len(task.all_entries) == 1
        assert len(task.accepted) == 1

        rsp = api_client.json_post('/tasks/execute/', data=json.dumps(payload))
        assert rsp.status_code == 200
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(OC.task_execution_results_schema, data)
        assert not errors

        task = self.get_task_queue(manager)
        task.execute()

        assert len(task.all_entries) == 1
        # Rejected due to Seen
        assert len(task.accepted) == 0

        # Forcing the entry not to be disabled
        entry['force'] = True

        rsp = api_client.json_post('/tasks/execute/', data=json.dumps(payload))
        assert rsp.status_code == 200
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(OC.task_execution_results_schema, data)
        assert not errors

        task = self.get_task_queue(manager)
        task.execute()

        assert len(task.all_entries) == 1
        assert len(task.accepted) == 1

    def test_inject_with_fields(self, api_client, manager, schema_match):
        fields = {'imdb_id': "tt1234567", 'tmdb_id': "1234567"}
        entry = {'title': "injected", 'url': 'http://test.com', 'fields': fields, 'accept': True}

        payload = {"inject": [entry], 'tasks': ['test_task']}

        rsp = api_client.json_post('/tasks/execute/', data=json.dumps(payload))
        assert rsp.status_code == 200
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(OC.task_execution_results_schema, data)
        assert not errors

        task = self.get_task_queue(manager)
        task.execute()

        assert len(task.all_entries) == 1
        assert len(task.accepted) == 1

        entry = task.find_entry(title='injected')
        assert entry['imdb_id'] == "tt1234567"
        assert entry['tmdb_id'] == "1234567"

    def test_multiple_entries(self, api_client, manager, schema_match):
        entry1 = {'title': "entry1", 'url': 'http://test.com', 'accept': True}
        entry2 = {'title': "entry2", 'url': 'http://test.com', 'accept': True}

        payload = {"inject": [entry1, entry2], 'tasks': ['test_task']}
        rsp = api_client.json_post('/tasks/execute/', data=json.dumps(payload))
        assert rsp.status_code == 200
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(OC.task_execution_results_schema, data)
        assert not errors

        task = self.get_task_queue(manager)
        task.execute()

        assert len(task.all_entries) == 2
        assert len(task.accepted) == 2

    def test_2nd_endpoint(self, api_client, manager, schema_match):
        entry = {'title': "injected", 'url': 'http://test.com', 'accept': True}

        payload = {"inject": [entry], 'tasks': ['test_task']}
        rsp = api_client.json_post('/inject/', data=json.dumps(payload))
        assert rsp.status_code == 200
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(OC.task_execution_results_schema, data)
        assert not errors

        task = self.get_task_queue(manager)
        task.execute()

        assert len(task.all_entries) == 1
        assert len(task.accepted) == 1


class TestExecuteMultipleTasks:
    config = """
        tasks:
          test_task1:
            mock:
              - title: accept_me1
            accept_all: yes
          test_task2:
            mock:
              - title: accept_me2
            accept_all: yes
        """

    def test_execute_multiple_tasks(self, api_client, manager, schema_match):
        rsp = api_client.json_post('/tasks/execute/', data=json.dumps({}))
        assert rsp.status_code == 422
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(base_message, data)
        assert not errors

        payload = {'tasks': ['non_existing_test_task']}
        rsp = api_client.json_post('/tasks/execute/', data=json.dumps(payload))
        assert rsp.status_code == 404
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(base_message, data)
        assert not errors

        payload = {'tasks': ['test_task1', 'test_task2']}
        rsp = api_client.json_post('/tasks/execute/', data=json.dumps(payload))
        assert rsp.status_code == 200
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(OC.task_execution_results_schema, data)
        assert not errors
